###############################################################################
#
# Crossbar.io Master
# Copyright (c) Crossbar.io Technologies GmbH. Licensed under EUPLv1.2.
#
###############################################################################

import uuid

from autobahn import wamp
from autobahn.wamp.types import CallDetails, PublishOptions
from autobahn.wamp.exception import ApplicationError

from cfxdb.user import Organization


class UserManager(object):
    """
    User management API (domain-global).

    Prefix: ``crossbarfabriccenter.user.``
    """

    _PUBOPTS = PublishOptions(acknowledge=True)

    def __init__(self, session, db, schema):
        self._session = session
        self.log = session.log
        self.db = db
        self.schema = schema
        self._prefix = None

    def register(self, session, prefix, options):
        self._prefix = prefix
        return session.register(self, prefix=prefix, options=options)

    @wamp.register(None)
    def list_organizations(self, details=None):
        """
        List all organizations accessible for the calling user. An organization
        is accessible for a user when at least one role was granted.

        :procedure: ``crossbarfabriccenter.user.list_organizations``

        Here is an example:

        .. code-block:: python

            async def test_list_organizations(session):
                oids = await session.call('crossbarfabriccenter.organization.list_organizations')
                print('got organizations:', oids)
                return oids

        :param details: Call details
        :type details: :class:`autobahn.wamp.types.CallDetails`

        :returns: List of OIDs of management realms
        :rtype: list[str]
        """
        assert details is None or isinstance(details, CallDetails)

        self.log.info('{klass}.list_organizations(details={details})', klass=self.__class__.__name__, details=details)

        with self.db.begin() as txn:
            org_oids = self.schema.organizations.select(txn, return_keys=True, return_values=False)
            if org_oids:
                # we now have a list of uuid.UUID objects: convert to strings
                return [str(oid) for oid in org_oids]
            else:
                return []

    @wamp.register(None)
    def get_organization(self, org_oid, details=None):
        """
        Return the organization definition by object OID. When no such object exists,
        an error will be returned.

        :procedure: ``crossbarfabriccenter.user.get_organization``
        :error: ``crossbar.error.no_such_object``

        Here is a typical example, fetching all organization (accessible by the user):

        .. code-block:: python

            async def test_get_all_orgs(session):
                oids = await session.call('crossbarfabriccenter.user.get_organizations')
                orgs = []
                if oids:
                    for oid in oids:
                        org = await session.call('crossbarfabriccenter.user.get_organization', oid)
                        print('got organization {}: {}'.format(org.oid, org))
                        orgs.append(org)
                return orgs

        :param org_oid: OID of the organization to get
        :type org_oid: str

        :param details: Call details
        :type details: :class:`autobahn.wamp.types.CallDetails`

        :returns: The organization database object (marshaled instance)
        :rtype: :class:`cfxdb.user.Organization`
        """
        assert type(org_oid) == str
        assert details is None or isinstance(details, CallDetails)

        self.log.info('{klass}.get_organization(org_oid={org_oid}, details={details})',
                      org_oid=org_oid,
                      klass=self.__class__.__name__,
                      details=details)

        org_oid = uuid.UUID(org_oid)

        with self.db.begin() as txn:

            org = self.schema.organizations[txn, org_oid]
            if org:
                self.log.info('Organization loaded:\n{org}', org=org)
                return org.marshal()
            else:
                raise ApplicationError('crossbar.error.no_such_object', 'no organization with oid {}'.format(org_oid))

    @wamp.register(None)
    async def create_organization(self, organization, details=None):
        """
        Create and store a new organization.

        :param organization:
        :param details: Call details
        :type details: :class:`autobahn.wamp.types.CallDetails`
        """
        assert type(organization) == dict
        assert details is None or isinstance(details, CallDetails)

        self.log.info('{klass}.create_organization(organization={organization}, details={details})',
                      klass=self.__class__.__name__,
                      organization=organization,
                      details=details)

        obj = Organization.parse(organization)
        obj.oid = uuid.uuid4()

        with self.db.begin(write=True) as txn:
            self.schema.organizations[txn, obj.oid] = obj

        self.log.info('new Organization object stored in database:\n{obj}', obj=obj)

        res_obj = obj.marshal()

        await self._session.publish('{}on_organization_created'.format(self._prefix), res_obj, options=self._PUBOPTS)

        self.log.info('Management API event <on_organization_created> published:\n{res_obj}', res_obj=res_obj)

        return res_obj

    @wamp.register(None)
    async def modify_organization(self, org_oid, org_delta, details=None):
        """
        Modify an existing organization.

        :param org_oid:
        :param org_delta:
        :param details: Call details
        :type details: :class:`autobahn.wamp.types.CallDetails`
        """
        raise NotImplementedError()

    @wamp.register(None)
    async def delete_organization(self, org_oid, cascade=False, details=None):
        """
        Delete an organization.

        :param org_oid:
        :param cascade:
        :param details: Call details
        :type details: :class:`autobahn.wamp.types.CallDetails`
        """
        assert type(org_oid) == str
        assert details is None or isinstance(details, CallDetails)

        self.log.info('{klass}.delete_organization(details={details})', klass=self.__class__.__name__, details=details)

        try:
            oid = uuid.UUID(org_oid)
        except Exception as e:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid oid "{}"'.format(str(e)))

        with self.db.begin(write=True) as txn:
            obj = self.schema.organizations[txn, oid]
            if obj:
                del self.schema.organizations[txn, oid]
            else:
                raise ApplicationError('crossbar.error.no_such_object', 'no object with oid {} found'.format(oid))

        self.log.info('Organization object deleted from database:\n{obj}', obj=obj)

        res_obj = obj.marshal()

        await self._session.publish('{}on_organization_deleted'.format(self._prefix), res_obj, options=self._PUBOPTS)

        return res_obj

    @wamp.register(None)
    def get_user(self, user_id, details=None):
        """
        Get a particular user by OID.

        :param user_id: OID of the user object to retrieve.
        :type user_id: str

        :param details: Call details
        :type details: :class:`autobahn.wamp.types.CallDetails`
        """
        assert type(user_id) == str
        assert details is None or isinstance(details, CallDetails)

        self.log.info('{klass}.get_user(user_id={user_id}, details={details})',
                      user_id=user_id,
                      klass=self.__class__.__name__,
                      details=details)

        user_id = uuid.UUID(user_id)

        with self.db.begin() as txn:

            user = self.schema.users[txn, user_id]
            if user:
                return user.marshal()
            else:
                raise ApplicationError('crossbar.error.no_such_object', 'no user with oid {}'.format(user_id))

    @wamp.register(None)
    def get_user_by_pubkey(self, pubkey, details=None):
        """
        Get a user by public key.

        :param pubkey: Public key of the user to retrieve.
        :type pubkey: str

        :param details: Call details
        :type details: :class:`autobahn.wamp.types.CallDetails`
        """
        assert type(pubkey) == str
        assert details is None or isinstance(details, CallDetails)

        with self.db.begin() as txn:
            user_id = self.schema.idx_users_by_pubkey[txn, pubkey]
            if user_id:
                user = self.schema.users[txn, user_id]
                assert user
                return user.marshal()
            else:
                raise ApplicationError('crossbar.error.no_such_object', 'no user with pubkey {}'.format(pubkey))

    @wamp.register(None)
    def get_user_by_email(self, email, details=None):
        """

        :param email: Email (name) of the user to retrieve.

        :param details: Call details
        :type details: :class:`autobahn.wamp.types.CallDetails`
        """
        assert type(email) == str
        assert details is None or isinstance(details, CallDetails)

        with self.db.begin() as txn:
            user_id = self.schema.idx_users_by_email[txn, email]
            if user_id:
                user = self.schema.users[txn, user_id]
                assert user
                return user.marshal()
            else:
                raise ApplicationError('crossbar.error.no_such_object', 'no user with email "{}"'.format(email))

    @wamp.register(None)
    async def modify_user(self, user_id, user_delta, details=None):
        """
        Modify an existing user.

        :procedure: ``crossbarfabriccenter.user.modify_user``
        :event: ``crossbarfabriccenter.user.on_user_modified``
        :error: ``crossbar.error.no_such_object``

        :param user_id: The user object to modify.
        :type user_id: str

        :param user_delta: The modification object with attributes present for values to be modified.
        :type user_delta: marshaled instance of :class:`cfxdb.user.User`

        :param details: Call details
        :type details: :class:`autobahn.wamp.types.CallDetails`

        :returns: A change object which with attributes and values present the reflect
            the actual modification that was made.
        :rtype: marshaled instance of :class:`cfxdb.user.User`
        """
        raise NotImplementedError()

    @wamp.register(None)
    def list_users(self, details=None):
        """
        List users.

        :param details: Call details
        :type details: :class:`autobahn.wamp.types.CallDetails`
        """
        assert details is None or isinstance(details, CallDetails)

        self.log.info('{klass}.list_users(details={details})', klass=self.__class__.__name__, details=details)

        with self.db.begin() as txn:
            users = [str(user_id) for user_id in self.schema.users.select(txn, return_values=False)]
            return users

    @wamp.register(None)
    async def list_users_by_organization(self, org_id, details=None):
        """
        List users with at least one role assigned on an organization.

        :param org_id: The organization to list users for.
        :type org_id: str

        :param details: Call details
        :type details: :class:`autobahn.wamp.types.CallDetails`
        """
        raise NotImplementedError()

    @wamp.register(None)
    async def list_organizations_by_user(self, user_id, details=None):
        """
        List organizations with roles assigned for a given user.

        :param user_id: The user to list organizations for.
        :type user_id: str

        :param details: Call details
        :type details: :class:`autobahn.wamp.types.CallDetails`
        """
        raise NotImplementedError()

    @wamp.register(None)
    async def set_roles_on_organization_for_user(self, org_id, user_id, roles, details=None):
        """
        Set roles for a user on an organization.

        :param org_id:
        :type org_id: str

        :param user_id:
        :type user_id: str

        :param roles:
        :type roles: list[str]

        :param details: Call details
        :type details: :class:`autobahn.wamp.types.CallDetails`
        """
        raise NotImplementedError()

    @wamp.register(None)
    async def get_roles_on_organization_for_user(self, org_id, user_id, details=None):
        """
        Get roles for a user on an organization.

        :param org_id:
        :type org_id: str

        :param user_id:
        :type user_id: str

        :param details: Call details
        :type details: :class:`autobahn.wamp.types.CallDetails`

        :returns: Roles a user currently has on the organization.
        :rtype: list[str]
        """
        raise NotImplementedError()



###############################################################################
#
# Crossbar.io Master
# Copyright (c) Crossbar.io Technologies GmbH. Licensed under EUPLv1.2.
#
###############################################################################

import re
import six
import uuid
from datetime import datetime

from autobahn import wamp
from autobahn.util import utcnow, utcstr
from autobahn.wamp.types import PublishOptions, CallDetails
from autobahn.wamp.exception import ApplicationError

from crossbar.master.node.authenticator import Authenticator

from cfxdb.mrealm import ManagementRealm, Node
from cfxdb.user import UserRole, UserMrealmRole

# users cannot use realm names starting with these:
_PROTECTED_REALM_PREFIXES = [
    'io.crossbar',
    'com.crossbario',
    'de.crossbario',
    'crossbar',
    'crossbar-wamp',
    'crossbarfabric',
    'crossbarfabric-wamp',
    'crossbarfabriccenter',
    'crossbario',
    'fabric',
    'autobahn',
    'wamp',
    'ws.wamp',
    'io.wamp',
]

ERROR_INVALID_REALM = 'fabric.invalid-realm'
ERROR_REALM_ALREADY_EXISTS = 'fabric.realm-already-exists'
ERROR_NO_SUCH_REALM = 'fabric.no-such-realm'
ERROR_NOT_AUTHORIZED = 'fabric.not-authorized'
ERROR_NODE_ALREADY_PAIRED = 'fabric.node-already-paired'
ERROR_NODE_NOT_PAIRED = 'fabric.node-not-paired'
ERROR_NO_SUCH_NODE = 'fabric.node-not-exist'

_REALM_NAME_PAT_STR = r"^[A-Za-z][A-Za-z0-9_\-@\.]{2,254}$"
_REALM_NAME_PAT = re.compile(_REALM_NAME_PAT_STR)

ON_REALM_ASSIGNED = 'crossbarfabriccenter.mrealm.on_realm_created'


def _check_realm_name(name):
    """
    Check a realm name.

    Valid (management) realm names in Crossbar.io should be
    a strict subset of what is allowed in Crossbar.io (Community).

    Hence, the following code needs to be in line with the code in

        crossbar.common.checkconfig.check_realm_name
    """
    if not isinstance(name, six.text_type):
        msg = 'invalid realm name "{}" - type must be string, was {}'.format(name, type(name))
        raise ApplicationError(ERROR_INVALID_REALM, msg)
    if not _REALM_NAME_PAT.match(name):
        msg = 'invalid realm name "{}" - must match regular expression {}'.format(name, _REALM_NAME_PAT_STR)
        raise ApplicationError(ERROR_INVALID_REALM, msg)
    for prefix in _PROTECTED_REALM_PREFIXES:
        if name.startswith(prefix):
            msg = 'invalid realm name "{}": names starting with "{}" are protected'.format(name, prefix)
            raise ApplicationError(ERROR_INVALID_REALM, msg)


def _rtype(realm_type):
    if realm_type == 'realm':
        return ManagementRealm.RTYPE_MREALM
    elif realm_type == 'app':
        return ManagementRealm.RTYPE_APP
    else:
        raise ValueError('invalid realm_type "{}"'.format(realm_type))


class MrealmManager(object):
    """
    Management Realms API (domain-global).

    Prefix: ``crossbarfabriccenter.mrealm.``
    """
    def __init__(self, session, db, schema):
        """

        :param session: crossbar.master.node.controller.DomainController
        :param db:
        :param schema:
        """
        self.log = session.log
        self.db = db
        self.schema = schema

        # DomainController._domain_mgr
        # DomainController._user_mgr
        # DomainController._mrealm_mgr
        self._session = session

    def register(self, session, prefix, options):
        return session.register(self, prefix=prefix, options=options)

    @wamp.register(None)
    def list_mrealms(self, return_names=None, details=None):
        """
        List the management realms accessible for the calling user. A mrealm is accessible for a user
        when at least one role was granted.

        :procedure: ``crossbarfabriccenter.mrealm.list_mrealms``

        Here is an example:

        .. code-block:: python

            async def test_list_mrealms(session):
                oids = await session.call('crossbarfabriccenter.mrealm.list_mrealms')
                print('got realms:', oids)
                return oids

        :param return_names: Return realm names instead of  object IDs
        :type return_names: bool

        :param details: Call details
        :type details: :class:`autobahn.wamp.types.CallDetails`

        :returns: List of OIDs (or names) of management realms
        :rtype: list[str]
        """
        assert return_names is None or type(return_names) == bool
        assert details is None or isinstance(details, CallDetails)

        self.log.info('{klass}.list_mrealms(return_names={return_names}, details={details})',
                      klass=self.__class__.__name__,
                      return_names=return_names,
                      details=details)

        if details.caller_authrole != Authenticator.GLOBAL_USER_REALM_USER_ROLE:
            raise Exception('proc not implemented for caller role "{}"'.format(details.caller_authrole))

        if return_names:
            with self.db.begin() as txn:
                realms = self.schema.mrealms.select(txn, return_keys=False, return_values=True)
                res = sorted([realm.name for realm in realms])
        else:
            with self.db.begin() as txn:
                realm_oids = self.schema.mrealms.select(txn, return_keys=True, return_values=False)
                res = [str(realm_oid) for realm_oid in realm_oids]
        return res

    @wamp.register(None)
    def get_mrealm_by_name(self, mrealm_name, details=None):
        """

        :param mrealm_name:

        :param details: Call details
        :type details: :class:`autobahn.wamp.types.CallDetails`

        :return:
        """
        assert type(mrealm_name) == str
        assert details is None or isinstance(details, CallDetails)

        self.log.info('{klass}.get_mrealm_by_name(mrealm_oid={mrealm_name}, details={details})',
                      mrealm_name=mrealm_name,
                      klass=self.__class__.__name__,
                      details=details)

        if details.caller_authrole != Authenticator.GLOBAL_USER_REALM_USER_ROLE:
            raise Exception('proc not implemented for caller role "{}"'.format(details.caller_authrole))

        with self.db.begin() as txn:

            mrealm_oid = self.schema.idx_mrealms_by_name[txn, mrealm_name]
            if mrealm_oid:
                mrealm = self.schema.mrealms[txn, mrealm_oid]
                self.log.info('Management realm loaded:\n{mrealm}', mrealm=mrealm)
                return mrealm.marshal()
            else:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no management realm with name {}'.format(mrealm_name))

    @wamp.register(None)
    def get_mrealm(self, mrealm_oid, details=None):
        """
        Return the management realm definition by object OID. When no such object exists,
        an error will be returned.

        :procedure: ``crossbarfabriccenter.mrealm.get_mrealm``
        :error: ``crossbar.error.no_such_object``

        Here is a typical example, fetching all management realms:

        .. code-block:: python

            async def test_get_all_mrealms(session):
                oids = await session.call('crossbarfabriccenter.mrealm.list_mrealms')
                mrealms = []
                if oids:
                    for oid in oids:
                        mrealm = await session.call('crossbarfabriccenter.mrealm.get_mrealm', oid)
                        print('got realm {}: {}'.format(mrealm.oid, mrealm))
                        mrealms.append(mrealm)
                return mrealms

        :param mrealm_oid: OID of the management realm to get
        :type mrealm_oid: str

        :param details: Call details
        :type details: :class:`autobahn.wamp.types.CallDetails`

        :returns: The management realm (marshaled instance)
        :rtype: :class:`cfxdb.mrealm.ManagementRealm`
        """
        assert type(mrealm_oid) == str
        assert details is None or isinstance(details, CallDetails)

        self.log.info('{klass}.get_realm(mrealm_oid={mrealm_oid}, details={details})',
                      mrealm_oid=mrealm_oid,
                      klass=self.__class__.__name__,
                      details=details)

        if details.caller_authrole != Authenticator.GLOBAL_USER_REALM_USER_ROLE:
            raise Exception('proc not implemented for caller role "{}"'.format(details.caller_authrole))

        mrealm_oid = uuid.UUID(mrealm_oid)

        with self.db.begin() as txn:

            mrealm = self.schema.mrealms[txn, mrealm_oid]
            if mrealm:
                self.log.debug('Management realm loaded:\n{mrealm}', mrealm=mrealm)
                return mrealm.marshal()
            else:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no management realm with oid {}'.format(mrealm_oid))

    @wamp.register(None)
    async def create_mrealm(self, mrealm, details=None):
        """
        Create a new management realm from the creation object provided with
        partially present attributes (non-empty values).

        Create attribute ``name`` must be given, and ``label``, ``description`` and ``tags`` may be
        optionally set. The name must be unique within the domain and only consist of lower case
        latin letters, digits and the ``_`` character (and begin with a letter). You can use the ``label``,
        ``description`` and ``tags`` attributes for arbitrary text.

        .. note::
            A management realm is internally identified by an OID (a 128 bit UUID) which never changes,
            but the management *name* can be changed later (though this is considered an invasive action).
            You can use the ``label``, ``description`` and ``tags`` attributes for arbitrary text.

        Here is an example to create a new management realm:

        .. code-block:: python

            import random

            async def test_create_mrealm(session):
                name = 'my_mrealm_{}'.format(random.randint(0, 1000))
                obj = {'name', name}
                mrealm = await session.call('crossbarfabriccenter.mrealm.create_mrealm', obj)
                print('new mrealm {} created: {}'.format(mrealm['oid'], mrealm))
                return mrealm

        Keep in mind that a management realm ``name`` must be unique in the whole CFC domain!

        WAMP:

        :procedure: ``crossbarfabriccenter.mrealm.create_mrealm``
        :event: ``crossbarfabriccenter.mrealm.on_mrealm_created``
        :error: ``crossbar.error.already_exists``
        :error: ``crossbar.error.invalid_configuration``

        Signature:

        :param mrealm: Creation object for management realm.
        :type mrealm: dict marshaled from :class:`cfxdb.mrealm.ManagementRealm`

        :param details: Call details
        :type details: :class:`autobahn.wamp.types.CallDetails`

        :returns: Newly created management realm, including OID generated.
        :rtype: dict marshaled from :class:`cfxdb.mrealm.ManagementRealm`
        """
        assert type(mrealm) == dict
        assert details is None or isinstance(details, CallDetails)

        arg_mrealm = ManagementRealm.parse(mrealm)
        new_mrealm = None

        self.log.info('{klass}.create_realm(name={name}, details={details})',
                      name=arg_mrealm.name,
                      klass=self.__class__.__name__,
                      details=details)

        if details.caller_authrole != Authenticator.GLOBAL_USER_REALM_USER_ROLE:
            raise Exception('proc not implemented for caller role "{}"'.format(details.caller_authrole))

        # check for valid mrealm name
        _check_realm_name(arg_mrealm.name)

        with self.db.begin(write=True) as txn:
            # get calling user
            user_oid = self.schema.idx_users_by_email[txn, details.caller_authid]
            if not user_oid:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no user with authid "{}" exists'.format(details.caller_authid))

            mrealm_oid = self.schema.idx_mrealms_by_name[txn, arg_mrealm.name]
            if mrealm_oid:
                raise ApplicationError('crossbar.error.already_exists',
                                       'management realm with name "{}" already exists'.format(arg_mrealm.name))

            new_mrealm = ManagementRealm()
            new_mrealm.oid = uuid.uuid4()
            new_mrealm.label = arg_mrealm.label
            new_mrealm.description = arg_mrealm.description
            new_mrealm.tags = arg_mrealm.tags
            new_mrealm.name = arg_mrealm.name
            new_mrealm.created = datetime.utcnow()
            new_mrealm.owner = user_oid
            new_mrealm.cf_router = 'cfrouter1'
            new_mrealm.cf_container = 'cfcontainer1'  # FIXME: dynamic placement of mrealm

            # store new management realm
            self.schema.mrealms[txn, new_mrealm.oid] = new_mrealm

            # store roles for user that created the management realm
            roles = UserMrealmRole([UserRole.OWNER, UserRole.ADMIN, UserRole.USER, UserRole.GUEST])
            self.schema.users_mrealm_roles[txn, (user_oid, new_mrealm.oid)] = roles

        self.log.debug('Management realm stored:\n{mrealm}', mrealm=new_mrealm)

        mrealm_obj = new_mrealm.marshal()

        started = await self._session.config.controller.call('crossbar.activate_realm', mrealm_obj)
        self.log.info('Management realm started: \n{started}', started=started)

        await self._session.publish('crossbarfabriccenter.mrealm.on_mrealm_created',
                                    mrealm_obj,
                                    options=PublishOptions(acknowledge=True))

        self.log.debug('Management API event <on_realm_created> published:\n{mrealm_obj}', mrealm_obj=mrealm_obj)

        return mrealm_obj

    @wamp.register(None)
    def modify_mrealm(self, mrealm_oid, mrealm_diff, details=None):
        """
        Modify an existing management realm from the modification object provided with
        partially present attributes (non-empty values).

        Change attributes can include ``name``, ``label``, ``description`` and ``tags``.

        :procedure: ``crossbarfabriccenter.mrealm.modify_mrealm``
        :event: ``crossbarfabriccenter.mrealm.on_mrealm_modified``
        :error: ``crossbar.error.no_such_object``

        :param mrealm_oid: OID of management realm to modify.
        :type mrealm_oid: str

        :param mrealm_diff: modification object with partially present, changed attributes (OID of the object itself
            cannot be changed)
        :type mrealm_diff: marshaled instance of :class:`cfxdb.mrealm.ManagementRealm`

        :param details: Call details
        :type details: :class:`autobahn.wamp.types.CallDetails`

        :returns: modification object with effectively changed attributes and values
        :rtype: marshaled instance of :class:`cfxdb.mrealm.ManagementRealm`
        """
        assert details is None or isinstance(details, CallDetails)

        raise NotImplementedError()

    @wamp.register(None)
    async def delete_mrealm(self, mrealm_oid, cascade=False, details=None):
        """
        Delete a management realm. A management realm can only be deleted by the owner of the mrealm,
        and no object (such as paired nodes or assigned users) must relate to the mrealm anymore.

        :procedure: ``crossbarfabriccenter.mrealm.delete_mrealm``
        :event: ``crossbarfabriccenter.mrealm.on_mrealm_delete``
        :error: ``crossbar.error.no_such_object``

        Here is an example of deleting a management realm:

        .. code-block:: python

            async def test_delete_mrealm(session, mrealm):
                await session.call('crossbarfabriccenter.mrealm.delete_mrealm', mrealm.oid)

        Here is an example listening for management realm deleted events:

        .. code-block:: python

            async def test_on_mrealm_deleted(session):

                def on_delete(deleted):
                    print('mrealm {} deleted'.format(deleted.oid)

                await session.subscribe(on_delete, 'crossbarfabriccenter.mrealm.on_mrealm_deleted')
                print('listening to on_mrealm_deleted ..')

        :param mrealm_oid: OID of the management realm to be deleted.
        :type mrealm_oid: str

        :param cascade: Automatically unpair (but not delete) any nodes currently paired with and
            unassign (but not delete) any users currently assigned to the management realm to be deleted.
        :type cascade: bool

        :param details: Call details
        :type details: :class:`autobahn.wamp.types.CallDetails`

        :returns: deleted object
        :rtype: marshaled instance of :class:`cfxdb.mrealm.ManagementRealm`
        """
        self.log.info('{klass}.delete_realm(mrealm_oid={mrealm_oid}, details={details})',
                      klass=self.__class__.__name__,
                      mrealm_oid=mrealm_oid,
                      details=details)

        if details.caller_authrole != Authenticator.GLOBAL_USER_REALM_USER_ROLE:
            raise Exception('proc not implemented for caller role "{}"'.format(details.caller_authrole))

        if isinstance(mrealm_oid, str):
            mrealm_oid = uuid.UUID(mrealm_oid)

        with self.db.begin() as txn:
            mrealm = self.schema.mrealms[txn, mrealm_oid]
            if not mrealm:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no management realm with ID "{}"'.format(mrealm_oid))

            # FIXME: check access to mrealm given resource-level access control system
            # FIXME: complete cascade delete:
            paired_nodes = []
            for node_id in self.schema.idx_nodes_by_authid.select(txn,
                                                                  from_key=(mrealm_oid, ''),
                                                                  to_key=(uuid.UUID(int=mrealm_oid.int + 1), ''),
                                                                  return_keys=False):

                if not cascade:
                    raise ApplicationError(
                        'crossbar.error.dependent_objects',
                        'cannot delete management - dependent object exists and cascade not set: paired node {}'.
                        format(node_id))
                paired_nodes.append(node_id)

        # cascade delete: unpair any nodes currently paired with the mrealm being deleted
        unpaired_nodes = []
        if cascade:
            for node_id in paired_nodes:
                unpaired_node = await self.unpair_node(str(node_id), details=details)
                unpaired_nodes.append(unpaired_node)
        else:
            assert not paired_nodes

        await self._session.config.controller.call('crossbar.deactivate_realm', mrealm.marshal())

        with self.db.begin(write=True) as txn:
            del self.schema.mrealms[txn, mrealm_oid]

        deleted = {
            'oid': str(mrealm_oid),
            'name': mrealm.name,
            'created': utcstr(mrealm.created),
            'deleted': utcnow(),
            'owner': str(mrealm.owner),
            'caller': details.caller_authid,
            'unpaired': [node['oid'] for node in unpaired_nodes],
        }

        await self._session.publish('crossbarfabriccenter.mrealm.on_realm_deleted',
                                    deleted,
                                    options=PublishOptions(acknowledge=True))

        return deleted

    @wamp.register(None)
    async def delete_mrealm_by_name(self, mrealm_name, cascade=False, details=None):
        """
        Delete an existing management realm (by name).

        :param mrealm_name: Name of the management realm to be deleted.
        :type mrealm_name: str

        :param cascade: Automatically unpair (but not delete) any nodes currently paired with and
            unassign (but not delete) any users currently assigned to the management realm to be deleted.
        :type cascade: bool

        :param details: Call details
        :type details: :class:`autobahn.wamp.types.CallDetails`

        :return:
        """
        self.log.info('{klass}.delete_mrealm_by_name(mrealm_name={mrealm_name}, cascade={cascade}, details={details})',
                      klass=self.__class__.__name__,
                      mrealm_name=mrealm_name,
                      cascade=cascade,
                      details=details)

        with self.db.begin() as txn:
            mrealm_oid = self.schema.idx_mrealms_by_name[txn, mrealm_name]

        if mrealm_oid:
            deleted = await self.delete_mrealm(mrealm_oid, cascade=cascade, details=details)
            return deleted
        else:
            raise ApplicationError('crossbar.error.no_such_object',
                                   'no management realm with name "{}"'.format(mrealm_name))

    @wamp.register(None)
    def set_roles_on_mrealm_for_user(self, mrealm_oid, user_oid, user_roles, details=None):
        """
        Set (user) roles on management realm for user. Only the owner of a management realm
        is allowed to set the roles of a user on that management realm.

        :procedure: ``crossbarfabriccenter.mrealm.set_roles_on_mrealm_for_user``

        :param mrealm_oid: OID of management realm on which to grant roles on.
        :type mrealm_oid: str

        :param user_oid: OID of user to which to grant roles to.
        :type user_oid: str

        :param roles: List of roles which to grant, from ``owner``, ``admin``, ``user``, ``guest``.
            The list of roles supplied supplied replaces any existing list of assigned roles.
            Supplying an empty list of roles (or ``None``) revokes any permissions and access for the user.
        :type roles: list[str]

        :param details: Call details
        :type details: :class:`autobahn.wamp.types.CallDetails`
        """
        raise NotImplementedError()

    @wamp.register(None)
    def list_nodes(self, details=None):
        """

        :procedure: ``crossbarfabriccenter.mrealm.list_nodes``

        :param details: Call details
        :type details: :class:`autobahn.wamp.types.CallDetails`
        """
        raise NotImplementedError()

    @wamp.register(None)
    def list_nodes_by_mrealm(self, mrealm_id, details=None):
        """

        :procedure: ``crossbarfabriccenter.mrealm.list_nodes_by_mrealm``

        :param mrealm_id:

        :param details: Call details
        :type details: :class:`autobahn.wamp.types.CallDetails`
        """
        raise NotImplementedError()

    @wamp.register(None)
    def list_nodes_by_mrealm_name(self, mrealm_name, details=None):
        """

        :procedure: ``crossbarfabriccenter.mrealm.list_nodes_by_mrealm_name``

        :param mrealm_name:

        :param details: Call details
        :type details: :class:`autobahn.wamp.types.CallDetails`
        """
        raise NotImplementedError()

    @wamp.register(None)
    def get_node(self, node_id, details=None):
        """

        :procedure: ``crossbarfabriccenter.mrealm.list_nodes_by_mrealm``

        :param node_id:

        :param details: Call details
        :type details: :class:`autobahn.wamp.types.CallDetails`
        """
        raise NotImplementedError()

    @wamp.register(None)
    def modify_node(self, node_id, node_delta, details=None):
        """

        :procedure: ``crossbarfabriccenter.mrealm.list_nodes_by_mrealm``

        :param node_id:
        :param node_delta:

        :param details: Call details
        :type details: :class:`autobahn.wamp.types.CallDetails`
        """
        raise NotImplementedError()

    @wamp.register(None)
    def delete_node(self, node_id, details=None):
        """

        :procedure: ``crossbarfabriccenter.mrealm.delete_node``

        :param node_id:

        :param details: Call details
        :type details: :class:`autobahn.wamp.types.CallDetails`
        """
        raise NotImplementedError()

    @wamp.register(None)
    def stat_node(self, node_id, details=None):
        """

        :procedure: ``crossbarfabriccenter.mrealm.stat_node``

        :param node_id:

        :param details: Call details
        :type details: :class:`autobahn.wamp.types.CallDetails`
        """
        raise NotImplementedError()

    @wamp.register(None)
    def get_node_by_name(self, mrealm_name, node_name, details=None):
        """

        :param mrealm_name:

        :param details: Call details
        :type details: :class:`autobahn.wamp.types.CallDetails`

        :return:
        """
        assert type(mrealm_name) == str
        assert details is None or isinstance(details, CallDetails)

        self.log.info('{klass}.get_mrealm_by_name(mrealm_oid={mrealm_name}, details={details})',
                      mrealm_name=mrealm_name,
                      klass=self.__class__.__name__,
                      details=details)

        if details.caller_authrole != Authenticator.GLOBAL_USER_REALM_USER_ROLE:
            raise Exception('proc not implemented for caller role "{}"'.format(details.caller_authrole))

        with self.db.begin() as txn:

            mrealm_oid = self.schema.idx_mrealms_by_name[txn, mrealm_name]
            if not mrealm_oid:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no management realm with name {}'.format(mrealm_name))

            node_oid = self.schema.idx_nodes_by_name[txn, (mrealm_oid, node_name)]
            if not node_oid:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no node named {} in management realm {}'.format(node_name, mrealm_name))

            node = self.schema.nodes[node_oid]
            if node.mrealm_oid != mrealm_oid:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no node named {} in management realm {}'.format(node_name, mrealm_name))

        return node.marshal()

    @wamp.register(None)
    async def pair_node(self, pubkey, realm_name, authid, authextra=None, details=None):
        """
        Pair a user Crossbar.io node with a user management realm.

        :procedure: ``crossbarfabriccenter.mrealm.pair_node``

        :param pubkey: The public key of the node to pair (HEX encoded).
        :type pubkey: str

        :param realm_name: The realm to which to pair the node to.
        :type realm_name: str

        :param node_id: The ID for the node to pair the node as.
        :type node_id: str

        :param authextra: Optional extra information handed out to the node
            when the node authenticates at CFC.
        :type authextra: None or dict

        :param details: Call details
        :type details: :class:`autobahn.wamp.types.CallDetails`

        :returns: Node paired information.
        :rtype: dict
        """
        self.log.info(
            '{klass}.pair_node(pubkey={pubkey}, realm_name={realm_name}, authid={authid}, authextra={authextra}, details={details})',
            pubkey=pubkey,
            realm_name=realm_name,
            authid=authid,
            authextra=authextra,
            klass=self.__class__.__name__,
            details=details)

        if details.caller_authrole != Authenticator.GLOBAL_USER_REALM_USER_ROLE:
            raise Exception('proc not implemented for caller role "{}"'.format(details.caller_authrole))

        node = None

        with self.db.begin(write=True) as txn:
            # get calling user
            user_oid = self.schema.idx_users_by_email[txn, details.caller_authid]
            if not user_oid:
                raise Exception('no such user')

            # check that the mrealm exists
            mrealm_oid = self.schema.idx_mrealms_by_name[txn, realm_name]
            mrealm = self.schema.mrealms[txn, mrealm_oid]
            if not mrealm_oid or not mrealm:
                msg = 'no realm named "{}" exists'.format(realm_name)
                raise ApplicationError(ERROR_NO_SUCH_REALM, msg)

            # owner of the mrealm
            owner = self.schema.users[txn, mrealm.owner]

            # check authorization of the calling user
            roles = self.schema.users_mrealm_roles[txn, (user_oid, mrealm_oid)]

            if not roles:
                msg = '"{}" not authorized to pair nodes to mrealm "{}" owned by {} (no roles assigned on mrealm)'.format(
                    details.caller_authid, realm_name, owner.email)
                raise ApplicationError(ERROR_NOT_AUTHORIZED, msg)

            # FIXME: allow UserRole.SUPERUSER (depends on: https://github.com/crossbario/cfxdb/issues/52)
            elif UserRole.OWNER not in roles.roles:
                msg = '"{}" not authorized to pair nodes to mrealm "{}" owned by {} (roles assigned on mrealm: {})'.format(
                    details.caller_authid, realm_name, owner.email, roles.roles)
                raise ApplicationError(ERROR_NOT_AUTHORIZED, msg)

            node_oid = self.schema.idx_nodes_by_pubkey[txn, pubkey]
            if not node_oid:
                self.log.info('Node pubkey not found, creating new node database object ..')
                node = Node()
                node.oid = uuid.uuid4()
                node.pubkey = pubkey
            else:
                self.log.info('Node pubkey found in database and loaded')
                node = self.schema.nodes[txn, node_oid]
                if node.mrealm_oid:
                    msg = 'node with given pubkey is already paired to a realm'
                    raise ApplicationError(ERROR_NODE_ALREADY_PAIRED, msg)

            node.owner_oid = user_oid
            node.mrealm_oid = mrealm_oid
            node.authid = authid
            node.authextra = authextra

            self.schema.nodes[txn, node.oid] = node

        assert node

        node_obj = node.marshal()

        topic = 'crossbarfabriccenter.mrealm.on_node_paired'
        self.log.debug('Publishing to <{topic}>: {payload} ..', topic=topic, payload=node_obj)
        await self._session.publish(topic, node_obj, options=PublishOptions(acknowledge=True))

        return node_obj

    @wamp.register(None)
    async def unpair_node(self, node_oid, details=None):
        """
        Unpair a user Crossbar.io node from a user management realm.

        :param node_oid:

        :param details: Call details
        :type details: :class:`autobahn.wamp.types.CallDetails`

        :return:
        """
        self.log.debug('{klass}.unpair_node(node_oid={node_oid}, details={details}) ..',
                       klass=self.__class__.__name__,
                       node_oid=node_oid,
                       details=details)

        if details.caller_authrole != Authenticator.GLOBAL_USER_REALM_USER_ROLE:
            raise Exception('proc not implemented for caller role "{}"'.format(details.caller_authrole))

        node = None
        node_oid = uuid.UUID(node_oid)
        mrealm = None

        with self.db.begin(write=True) as txn:
            # get calling user
            user_oid = self.schema.idx_users_by_email[txn, details.caller_authid]
            if not user_oid:
                raise Exception('no such user')

            node = self.schema.nodes[txn, node_oid]
            if not node:
                raise ApplicationError('crossbar.error.no_such_object', 'no node with oid {}'.format(node_oid))

            # FIXME
            # if node.owner_oid != details.caller_authid:
            #    raise Exception('not authorized')

            if not node.mrealm_oid:
                raise ApplicationError(ERROR_NODE_NOT_PAIRED,
                                       'cannot unpair node: node with id {} is currently not paired'.format(node_oid))

            mrealm = self.schema.mrealms[txn, node.mrealm_oid]
            assert mrealm

            node.mrealm_oid = None
            node.authid = None
            node.authextra = None

            self.schema.nodes[txn, node_oid] = node

        assert node

        # kill any management uplink sessions from the unpaired node
        #
        killed = None
        try:
            controller_session = self._session.config.controller

            # FIXME: hard-coded "cfrouter1" worker ID
            controller_proc = 'crossbar.worker.cfrouter1.kill_by_authid'

            killed = await controller_session.call(
                controller_proc,
                str(mrealm.name),
                str(node_oid),
                reason='wamp.close.auth-changed',
                message=
                'Authentication information or permissions changed for node with authid "{}" - killing all currently active sessions for this authid'
                .format(node_oid))
        except:
            self.log.failure()

        node_obj = node.marshal()
        node_obj['mrealm_oid'] = str(mrealm.oid)
        node_obj['killed'] = killed

        topic = 'crossbarfabriccenter.mrealm.on_node_unpaired'
        self.log.debug('Publishing to <{topic}>: {payload} ..', topic=topic, payload=node_obj)
        await self._session.publish(topic, node_obj, options=PublishOptions(acknowledge=True))

        if killed:
            self.log.info('{klass}.unpair_node(node_oid={node_oid}): unpaired (killed {killed} active sessions).',
                          klass=self.__class__.__name__,
                          node_oid=node_oid,
                          killed=len(killed))
        else:
            self.log.info('{klass}.unpair_node(node_oid={node_oid}): unpaired (no active sessions).',
                          klass=self.__class__.__name__,
                          node_oid=node_oid)

        return node_obj

    @wamp.register(None)
    async def unpair_node_by_pubkey(self, pubkey, details=None):
        """

        :param pubkey:

        :param details: Call details
        :type details: :class:`autobahn.wamp.types.CallDetails`

        :return:
        """
        self.log.info('{klass}.unpair_node_by_pubkey(pubkey={pubkey}, details={details})',
                      klass=self.__class__.__name__,
                      pubkey=pubkey,
                      details=details)

        with self.db.begin() as txn:
            node_oid = self.schema.idx_nodes_by_pubkey[txn, pubkey]
        if node_oid:
            unpaired = await self.unpair_node(str(node_oid), details=details)
            return unpaired
        else:
            raise ApplicationError('crossbar.error.no_such_object', 'no node with pubkey {}'.format(pubkey))

    @wamp.register(None)
    async def unpair_node_by_name(self, realm_name, authid, details=None):
        """

        :param realm_name:
        :param authid:

        :param details: Call details
        :type details: :class:`autobahn.wamp.types.CallDetails`

        :return:
        """
        self.log.info('{klass}.unpair_node_by_name(realm_name={realm_name}, authid={authid}, details={details})',
                      klass=self.__class__.__name__,
                      realm_name=realm_name,
                      authid=authid,
                      details=details)

        raise NotImplementedError()



###############################################################################
#
# Crossbar.io Master
# Copyright (c) Crossbar.io Technologies GmbH. Licensed under EUPLv1.2.
#
###############################################################################

from crossbar.master.api.remote import RemoteApi

__all__ = ('RemoteProxyApi', )


class RemoteProxyApi(RemoteApi):

    PREFIX = 'crossbarfabriccenter.remote.proxy.'

    PROCS = {
        # these are worker level procedures
        'worker': [
            'get_proxy_transports',
            'get_proxy_transport',
            'start_proxy_transport',
            'stop_proxy_transport',
            'get_web_transport_services',
            'get_web_transport_service',
            'start_web_transport_service',
            'stop_web_transport_service',
            'get_proxy_routes',
            'get_proxy_realm_route',
            'list_proxy_realm_routes',
            'start_proxy_realm_route',
            'stop_proxy_realm_route',
            'get_proxy_connections',
            'get_proxy_connection',
            'start_proxy_connection',
            'stop_proxy_connection',
        ],
    }

    EVENTS = {
        # these are worker level topics
        'worker': [
            'on_proxy_transport_starting',
            'on_proxy_transport_started',
            'on_proxy_transport_stopping',
            'on_proxy_transport_stopped',
        ]
    }



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################

import os
from pprint import pformat
from typing import Dict, Any, Optional, Tuple, Set

from twisted.internet.base import ReactorBase
from twisted.internet.defer import Deferred, inlineCallbacks, returnValue
from twisted.internet.error import DNSLookupError

from txaio import make_logger, as_future, time_ns

from autobahn.wamp import cryptosign

from autobahn import wamp
from autobahn import util
from autobahn.wamp import types
from autobahn.wamp import message
from autobahn.wamp.auth import create_authenticator
from autobahn.wamp.exception import ApplicationError, TransportLost, ProtocolError
from autobahn.wamp.role import RoleDealerFeatures, RoleBrokerFeatures
from autobahn.wamp.component import _create_transport
from autobahn.wamp.interfaces import ITransportHandler, IMessage
from autobahn.twisted.wamp import Session, ApplicationSession
from autobahn.twisted.component import _create_transport_factory, _create_transport_endpoint
from autobahn.twisted.component import Component

from crossbar.interfaces import IRealmContainer
from crossbar._util import hltype, hlid, hlval
from crossbar.node import worker
from crossbar.worker.controller import WorkerController
from crossbar.worker.transport import TransportController
from crossbar.common.key import _read_node_key
from crossbar.router.auth import PendingAuthWampCra, PendingAuthTicket, PendingAuthScram
from crossbar.router.auth import AUTHMETHOD_MAP
from crossbar.router.session import RouterSessionFactory
from crossbar.router.session import RouterFactory

try:
    from crossbar.router.auth import PendingAuthCryptosign, PendingAuthCryptosignProxy
except ImportError:
    PendingAuthCryptosign = None  # type: ignore
    PendingAuthCryptosignProxy = None  # type: ignore

__all__ = (
    'ProxyWorkerProcess',
    'ProxyController',
    'ProxyConnection',
    'ProxyRoute',
)

log = make_logger()


class ProxyWorkerProcess(worker.NativeWorkerProcess):

    TYPE = 'proxy'
    LOGNAME = 'Proxy'


class ProxyFrontendSession(object):
    """
    A router-side proxy session that handles incoming client
    connections.
    """
    # Note: "roles" come from self._router.attach() in non-proxy code
    ROLES = {
        'broker':
        RoleBrokerFeatures(
            publisher_identification=True,
            pattern_based_subscription=True,
            session_meta_api=True,
            subscription_meta_api=True,
            subscriber_blackwhite_listing=True,
            publisher_exclusion=True,
            subscription_revocation=True,
            event_retention=True,
            payload_transparency=True,
            payload_encryption_cryptobox=True,
        ),
        'dealer':
        RoleDealerFeatures(
            caller_identification=True,
            pattern_based_registration=True,
            session_meta_api=True,
            registration_meta_api=True,
            shared_registration=True,
            progressive_call_results=True,
            registration_revocation=True,
            payload_transparency=True,
            testament_meta_api=True,
            payload_encryption_cryptobox=True,
            call_canceling=True,
        )
    }

    log = make_logger()

    def __init__(self, router_factory):
        self._router_factory = router_factory
        self._controller = router_factory.worker
        self._reset()

    def _reset(self):
        # after the frontend connection is open, this will be the frontend transport
        self.transport = None

        # if we have a backend connection, it'll be here (and be a
        # Session instance)
        self._backend_session = None

        # pending session id (before the session is fully joined)
        self._pending_session_id = None

        # authenticated+joined session information
        self._session_id = None
        self._authid = None
        self._realm = None
        self._authid = None
        self._authrole = None
        self._authmethod = None
        self._authprovider = None
        self._authextra = None

        self._custom_authextra = {}

    @property
    def realm(self):
        return self._realm

    @property
    def authid(self):
        return self._authid

    @property
    def authrole(self):
        return self._authrole

    @property
    def authmethod(self):
        return self._authmethod

    @property
    def authprovider(self):
        return self._authprovider

    @property
    def authextra(self):
        return self._authextra

    def onOpen(self, transport):
        """
        Callback fired when transport is open. May run asynchronously. The transport
        is considered running and is_open() would return true, as soon as this callback
        has completed successfully.

        :param transport: The WAMP transport.
        :type transport: object implementing :class:`autobahn.wamp.interfaces.ITransport`
        """
        self.log.debug('{func}(transport={transport})', func=hltype(self.onOpen), transport=transport)
        self.transport = transport

        # transport configuration
        if hasattr(self.transport, 'factory') and hasattr(self.transport.factory, '_config'):
            self._transport_config = self.transport.factory._config
        else:
            self._transport_config = {}

        self._custom_authextra = {
            'x_cb_proxy_node': self._router_factory._node_id,
            'x_cb_proxy_worker': self._router_factory._worker_id,
            'x_cb_proxy_peer': str(self.transport.peer),
            'x_cb_proxy_pid': os.getpid(),
        }

        self.log.info('{func} proxy frontend session connected from peer {peer}',
                      func=hltype(self.onOpen),
                      peer=hlval(self.transport.transport_details.peer) if self.transport.transport_details else None)

    def onClose(self, wasClean):
        """
        Callback fired when the transport has been closed.

        :param wasClean: Indicates if the transport has been closed regularly.
        :type wasClean: bool
        """
        self.log.info('{func} proxy frontend session closed (wasClean={wasClean})',
                      func=hltype(self.onClose),
                      wasClean=wasClean)

        # actually, at this point, the backend session should already be gone, but better check!
        if self._backend_session:
            self._controller.unmap_backend(self, self._backend_session)
            self._backend_session = None

        # reset everything (even though this frontend protocol instance should not be reused anyways)
        self._reset()

    @inlineCallbacks
    def onMessage(self, msg):
        """
        Callback fired when a WAMP message was received. May run asynchronously. The callback
        should return or fire the returned deferred/future when it's done processing the message.
        In particular, an implementation of this callback must not access the message afterwards.

        :param msg: The WAMP message received.
        :type msg: object implementing :class:`autobahn.wamp.interfaces.IMessage`
        """
        self.log.debug('{func} proxy frontend session onMessage(msg={msg})', func=hltype(self.onMessage), msg=msg)
        if self._session_id is None:
            # no frontend session established yet, so we expect one of HELLO, ABORT, AUTHENTICATE

            # https://wamp-proto.org/_static/gen/wamp_latest.html#session-establishment
            if isinstance(msg, message.Hello):
                yield self._process_Hello(msg)

            # https://wamp-proto.org/_static/gen/wamp_latest.html#session-closing
            elif isinstance(msg, message.Abort):
                self.transport.send(
                    message.Abort(ApplicationError.AUTHENTICATION_FAILED, message='Proxy authentication failed'))

            # https://wamp-proto.org/_static/gen/wamp_latest.html#wamp-level-authentication
            elif isinstance(msg, message.Authenticate):
                self._process_Authenticate(msg)

            else:
                raise ProtocolError("Received {} message while proxy frontend session is not joined".format(
                    msg.__class__.__name__))

        else:
            # frontend session is established: process WAMP message

            if isinstance(msg, message.Hello) or isinstance(msg, message.Abort) or isinstance(
                    msg, message.Authenticate):
                raise ProtocolError("Received {} message while proxy frontend session is already joined".format(
                    msg.__class__.__name__))

            # https://wamp-proto.org/_static/gen/wamp_latest.html#session-closing
            elif isinstance(msg, message.Goodbye):
                if self._backend_session:
                    self._controller.unmap_backend(self, self._backend_session)
                    self._backend_session = None
                else:
                    self.log.warn('{func} frontend session left, but no active backend session to close!',
                                  func=hltype(self.onMessage))

                # complete the closing handshake (initiated by the client in this case) by replying with GOODBYE
                self.transport.send(message.Goodbye(message="Proxy session closing"))
            else:
                if self._backend_session is None or self._backend_session._transport is None:
                    raise TransportLost(
                        "Expected to relay {} message, but proxy backend session or transport is gone".format(
                            msg.__class__.__name__, ))
                else:
                    # if we have an active backend connection, forward the WAMP message ..
                    self._backend_session._transport.send(msg)

    def frontend_accepted(self, accept):
        # we have done authentication with the client; now we can connect to
        # the backend (and we wait to tell the client they're
        # welcome until we have actually connected to the
        # backend).
        self.log.info('{func} proxy frontend session accepted ({accept})',
                      func=hltype(self.frontend_accepted),
                      accept=accept)

        if (hasattr(self.transport, '_cbtid') and hasattr(self.transport.factory, '_cookiestore')
                and self.transport.factory._cookiestore):
            self.transport.factory._cookiestore.setAuth(self.transport._cbtid, accept.authid, accept.authrole,
                                                        accept.authmethod, accept.authextra, accept.realm)

        result = Deferred()

        @inlineCallbacks
        def backend_connected(backend: ProxyBackendSession):
            # bytestream-level transport to backend router worker connected.
            try:
                # first, wait for the WAMP-level transport to connect before starting to join
                yield backend._on_connect

                # node private key
                key = _read_node_key(self._controller._cbdir, private=False)

                # authid of the connecting backend (proxy service) session is this proxy node's ID
                backend_authid = self._controller.node_id

                # authmethods we announce to the backend router we connect to
                authmethods = list(backend._authenticators.keys())

                # get marshalled transport details for this proxy frontend session
                if self.transport.transport_details:
                    _td = self.transport.transport_details.marshal()
                else:
                    _td = None

                # authentication extra we transmit from the proxy to backend router worker
                authextra = {
                    # for WAMP-cryptosign authentication of the proxy frontend
                    # to the backend router
                    "pubkey": key['hex'],

                    # forward authentication credentials of the connecting client
                    #
                    # the following are the effective (realm, authid, authrole) under
                    # which the client (proxy frontend connection) was successfully
                    # authenticated (using the authmethod+authprovider)
                    "proxy_realm": accept.realm,
                    "proxy_authid": accept.authid,
                    "proxy_authrole": accept.authrole,
                    "proxy_authmethod": accept.authmethod,
                    "proxy_authprovider": accept.authprovider,

                    # this is the authextra returned from the proxy frontend authenticator
                    # "proxy_authextra": accept.authextra,

                    # these are the transport details from the proxy frontend session. this is picked
                    # up in PendingAuthAnonymousProxy.hello() and PendingAuthCryptosignProxy.hello()
                    "proxy_authextra": {
                        "transport": _td
                    },
                }

                self.log.info(
                    '{func} proxy backend session authenticating with authmethods={authmethods}, pubkey={pubkey}: '
                    'proxy_authid="{proxy_authid}", proxy_authrole="{proxy_authrole}", proxy_realm="{proxy_realm}"',
                    func=hltype(backend_connected),
                    authmethods=hlval(authmethods),
                    pubkey=hlval(authextra['pubkey']),
                    proxy_authid=hlid(authextra['proxy_authid']),
                    proxy_authrole=hlid(authextra['proxy_authrole']),
                    proxy_realm=hlid(authextra['proxy_realm']))

                # now join WAMP session, which might first start WAMP authentication (for authmethod "cryptosign-proxy")
                backend.join(accept.realm, authmethods=authmethods, authid=backend_authid, authextra=authextra)

                def backend_joined(session, details):
                    self.log.debug('{func} proxy backend session joined (backend_session_id={backend_session_id})',
                                   backend_session_id=hlid(details.session),
                                   backend_session=session,
                                   pending_session_id=self._pending_session_id,
                                   func=hltype(backend_joined))
                    # we're ready now! store and return the backend session
                    self._backend_session = session

                    # we set the frontend session ID to that of the backend session mapped for our frontend session ..
                    self._session_id = details.session
                    # .. NOT our (fake) pending session ID (generated in the proxy worker)
                    # self._session_id = self._pending_session_id

                    # credentials of the backend session mapped for our frontend session
                    self._realm = details.realm
                    self._authid = details.authid
                    self._authrole = details.authrole

                    # this is the authextra returned for the backend session mapped for our frontend session
                    self._authextra = details.authextra

                    # authentication method & provider are the requested (and succeeding) ones
                    self._authmethod = accept.authmethod
                    self._authprovider = accept.authprovider

                    result.callback(session)

                backend.on('join', backend_joined)
                yield backend._on_ready
                return backend
            except Exception as e:
                self.log.failure()
                result.errback(e)

        def backend_failed(fail):
            result.errback(fail)

        # map and connect bytestream-level transport to backend router worker
        backend_d = self._controller.map_backend(
            self,
            accept.realm,
            accept.authid,
            accept.authrole,
            accept.authextra,
        )
        backend_d.addCallback(backend_connected)
        backend_d.addErrback(backend_failed)

        return result

    def _forward(self, msg):
        # we received a message on the backend connection: forward to client over frontend connection
        if self.transport:
            self.transport.send(msg)
        else:
            # FIXME: can we improve this?
            # eg when the frontend client connection has closed before we had a chance to stop the backend and we still
            # receive eg a Result(request=23033, args=[] ..) from the backend.
            self.log.debug('Trying to forward a message to the client, but no frontend transport! [{msg}]', msg=msg)

    @inlineCallbacks
    def _process_Hello(self, msg):
        """
        We have received a Hello from the frontend client.

        Now we do any authentication necessary with them and connect
        to our backend.
        """
        self.log.debug('{func} proxy frontend session processing HELLO (msg={msg})',
                       func=hltype(self._process_Hello),
                       msg=msg)
        self._pending_session_id = util.id()
        self._goodbye_sent = False

        realm = msg.realm

        # allow "Personality" classes to add authmethods
        extra_auth_methods = self._controller.personality.EXTRA_AUTH_METHODS
        authmethods = list(extra_auth_methods.keys()) + (msg.authmethods or ['anonymous'])

        # if the client had a reassigned realm during authentication, restore it from the cookie
        if hasattr(self.transport, '_authrealm') and self.transport._authrealm:
            if 'cookie' in authmethods:
                realm = self.transport._authrealm  # noqa
                authextra = self.transport._authextra  # noqa
            elif self.transport._authprovider == 'cookie':
                # revoke authentication and invalidate cookie (will be revalidated if following auth is successful)
                self.transport._authmethod = None
                self.transport._authrealm = None
                self.transport._authid = None
                if hasattr(self.transport, '_cbtid'):
                    self.transport.factory._cookiestore.setAuth(self.transport._cbtid, None, None, None, None, None)
            else:
                pass  # TLS authentication is not revoked here

        # already authenticated, eg via HTTP-cookie or TLS-client-certificate authentication
        if self.transport._authid is not None and (self.transport._authmethod == 'trusted'
                                                   or self.transport._authprovider in authmethods):
            msg.realm = realm
            msg.authid = self.transport._authid
            msg.authrole = self.transport._authrole
            msg.authextra = self.transport._authextra

        details = types.HelloDetails(realm=realm,
                                     authmethods=authmethods,
                                     authid=msg.authid,
                                     authrole=msg.authrole,
                                     authextra=msg.authextra,
                                     session_roles=msg.roles,
                                     pending_session=self._pending_session_id)

        auth_config = self._transport_config.get('auth', None)

        # if authentication is _not_ configured, allow anyone to join as "anonymous"!
        if not auth_config:
            # we ignore any details.authid the client might have announced, and use
            # a cookie value or a random value
            if hasattr(self.transport, "_cbtid") and self.transport._cbtid:
                # if cookie tracking is enabled, set authid to cookie value
                authid = self.transport._cbtid
            else:
                # if no cookie tracking, generate a random value for authid
                authid = util.generate_serial_number()

            # FIXME: really forward any requested authrole?
            authrole = msg.authrole

            auth_config = {
                'anonymous': {
                    'type': 'static',
                    'authrole': authrole,
                    'authid': authid,
                }
            }
            self.log.warn(
                '{func} no authentication configured for proxy frontend session (using builtin anonymous '
                'access policy)',
                func=hltype(self._process_Hello))

        for authmethod in authmethods:
            # invalid authmethod
            if authmethod not in AUTHMETHOD_MAP and authmethod not in extra_auth_methods:
                self.transport.send(
                    message.Abort(ApplicationError.AUTHENTICATION_FAILED,
                                  message='authmethod "{}" not allowed'.format(authmethod)))
                return

            # authmethod is valid, but not configured: continue trying other authmethods the client is announcing
            if authmethod not in auth_config:
                continue

            # authmethod not available
            if authmethod not in AUTHMETHOD_MAP and authmethod not in extra_auth_methods:
                self.log.debug("{func} client requested valid, but unavailable authentication method {authmethod}",
                               func=hltype(self._process_Hello),
                               authmethod=authmethod)
                continue

            if authmethod == 'cookie' and hasattr(self.transport,
                                                  "_cbtid") and self.transport._cbtid and self.transport._authid:
                hello_result = types.Accept(realm=realm,
                                            authid=self.transport._authid,
                                            authrole=self.transport._authrole,
                                            authmethod=self.transport._authmethod,
                                            authprovider='cookie',
                                            authextra=self.transport._authextra)
                self.log.debug(
                    '{func}: authenticating client using cookie-authentication, hello_result={hello_result}',
                    func=hltype(self._process_Hello),
                    hello_result=hello_result)

            else:
                # create instance of authenticator using authenticator class for the respective authmethod
                authklass = extra_auth_methods[authmethod] if authmethod in extra_auth_methods else AUTHMETHOD_MAP[
                    authmethod]

                if authklass is None:
                    self.log.warn('{func}: skipping authenticator for authmethod "{authmethod}"',
                                  func=hltype(self._process_Hello),
                                  authmethod=hlval(authmethod))
                    self.log.warn()
                    continue

                self.log.debug('{func}: instantiating authenticator class {authklass} for authmethod "{authmethod}"',
                               func=hltype(self._process_Hello),
                               authklass=hltype(authklass),
                               authmethod=hlval(authmethod))
                self._pending_auth = authklass(
                    self._pending_session_id,
                    self.transport.transport_details,
                    self._controller,
                    auth_config[authmethod],
                )
                try:
                    # call into authenticator for processing the HELLO message
                    hello_result = yield as_future(self._pending_auth.hello, realm, details)
                except Exception as e:
                    self.log.failure()
                    self.transport.send(
                        message.Abort(ApplicationError.AUTHENTICATION_FAILED,
                                      message='Frontend connection accept failed ({})'.format(e)))
                    return
                self.log.debug('{func} authmethod "{authmethod}" completed with result={hello_result}',
                               func=hltype(self._process_Hello),
                               authmethod=hlval(authmethod),
                               hello_result=hello_result)

            # if the frontend session is accepted right away (eg when doing "anonymous" authentication), process the
            # frontend accept ..
            if isinstance(hello_result, types.Accept):
                try:
                    # get a backend session mapped to the incoming frontend session
                    session = yield self.frontend_accepted(hello_result)
                except Exception as e:
                    self.log.failure()
                    self.transport.send(
                        message.Abort(ApplicationError.AUTHENTICATION_FAILED,
                                      message='Frontend connection accept failed ({})'.format(e)))
                    return

                def _on_backend_joined(session, details):
                    # we now got everything! the frontend is authenticated, and a backend session is associated.
                    msg = message.Welcome(self._session_id,
                                          ProxyFrontendSession.ROLES,
                                          realm=details.realm,
                                          authid=details.authid,
                                          authrole=details.authrole,
                                          authmethod=hello_result.authmethod,
                                          authprovider=hello_result.authprovider,
                                          authextra=dict(details.authextra or {}, **self._custom_authextra))
                    self._backend_session = session
                    self.transport.send(msg)
                    self.log.debug(
                        '{func} proxy frontend session WELCOME: session_id={session}, session={session}, '
                        'details="{details}"',
                        func=hltype(self._process_Hello),
                        session_id=hlid(self._session_id),
                        session=self,
                        details=details)

                session.on('join', _on_backend_joined)

            # if the client is required to do an authentication message exchange, answer sending a CHALLENGE message
            elif isinstance(hello_result, types.Challenge):
                self.transport.send(message.Challenge(hello_result.method, extra=hello_result.extra))

            # if the client is denied right away, answer by sending an ABORT message
            elif isinstance(hello_result, types.Deny):
                self.transport.send(message.Abort(hello_result.reason, message=hello_result.message))

            else:
                # should not arrive here: internal (logic) error
                self.log.warn('{func} internal error: unexpected authenticator return type {rtype}',
                              rtype=hltype(hello_result),
                              func=hltype(self._process_Hello))
                self.transport.send(
                    message.Abort(ApplicationError.AUTHENTICATION_FAILED,
                                  message='internal error: unexpected authenticator return type {}'.format(
                                      type(hello_result))))
            return

        self.transport.send(message.Abort(ApplicationError.NO_AUTH_METHOD, message='no suitable authmethod found'))

    @inlineCallbacks
    def _process_Authenticate(self, msg):
        self.log.debug('{func} proxy frontend session process AUTHENTICATE (msg={msg})',
                       func=hltype(self._process_Authenticate),
                       msg=msg)
        if self._pending_auth:
            if isinstance(self._pending_auth, PendingAuthTicket) or \
               isinstance(self._pending_auth, PendingAuthWampCra) or \
               isinstance(self._pending_auth, PendingAuthCryptosign) or \
               isinstance(self._pending_auth, PendingAuthScram):
                auth_result = yield as_future(self._pending_auth.authenticate, msg.signature)
                self.log.debug('{func} processed pending authentication {pending_auth}: {authresult}',
                               func=hltype(self._process_Authenticate),
                               pending_auth=self._pending_auth,
                               authresult=auth_result)
                if isinstance(auth_result, types.Accept):
                    try:
                        session = yield self.frontend_accepted(auth_result)
                    except Exception as e:
                        self.log.failure()
                        self.transport.send(
                            message.Abort(ApplicationError.AUTHENTICATION_FAILED,
                                          message='Frontend connection accept failed ({})'.format(e)))
                    else:

                        def _on_backend_joined(session, details):
                            msg = message.Welcome(self._session_id,
                                                  ProxyFrontendSession.ROLES,
                                                  realm=details.realm,
                                                  authid=details.authid,
                                                  authrole=details.authrole,
                                                  authmethod=auth_result.authmethod,
                                                  authprovider=auth_result.authprovider,
                                                  authextra=dict(details.authextra or {}, **self._custom_authextra))
                            self._backend_session = session
                            self.transport.send(msg)
                            self.log.debug(
                                '{func} proxy frontend session WELCOME: session_id={session_id}, '
                                'session={session}, msg={msg}',
                                func=hltype(self._process_Authenticate),
                                session_id=hlid(self._session_id),
                                session=self,
                                msg=msg)

                        session.on('join', _on_backend_joined)
                elif isinstance(auth_result, types.Deny):
                    self.transport.send(message.Abort(auth_result.reason, message=auth_result.message))
                else:
                    # should not arrive here: logic error
                    self.log.warn('{func} internal error: unexpected authenticator return type {rtype}',
                                  rtype=hltype(auth_result),
                                  func=hltype(self._process_Authenticate))
                    self.transport.send(
                        message.Abort(ApplicationError.AUTHENTICATION_FAILED,
                                      message='internal error: unexpected authenticator return type {}'.format(
                                          type(auth_result))))
            else:
                # should not arrive here: logic error
                self.transport.send(
                    message.Abort(ApplicationError.AUTHENTICATION_FAILED,
                                  message='internal error: unexpected pending authentication'))
        else:
            # should not arrive here: client misbehaving!
            self.transport.send(
                message.Abort(ApplicationError.AUTHENTICATION_FAILED, message='no pending authentication'))


ITransportHandler.register(ProxyFrontendSession)


class ProxyBackendSession(Session):
    """
    This is a single WAMP session to the real backend service

    There is one of these for every client connection. (In the future,
    we could multiplex over a single backend connection -- for now,
    there's a backend connection per frontend client).
    """
    def onOpen(self, transport):
        self.log.debug('{func}(transport={transport})', func=hltype(self.onOpen), transport=transport)
        # instance of Frontend (frontend_session)
        self._frontend = transport._proxy_other_side
        self._on_connect = Deferred()
        self._on_ready = Deferred()
        return Session.onOpen(self, transport)

    def onConnect(self):
        """
        The base class will call .join() which we do NOT want to do;
        instead we await the frontend sending its hello and forward
        that along.
        """
        self.log.debug('{func}(): on_connect={on_connect}', func=hltype(self.onConnect), on_connect=self._on_connect)
        self._on_connect.callback(None)

    def onChallenge(self, challenge):
        self.log.debug('{func}(challenge={})', func=hltype(self.onChallenge), challenge=challenge)
        if challenge.method == "cryptosign-proxy":
            return super(ProxyBackendSession, self).onChallenge(types.Challenge("cryptosign", extra=challenge.extra))

        return super(ProxyBackendSession, self).onChallenge(challenge)

    def onWelcome(self, _msg):
        self.log.debug('{func}(message={msg})', func=hltype(self.onWelcome), msg=_msg)
        # This is WRONG:
        # if msg.authmethod == "cryptosign-proxy":
        #     msg.authmethod = "cryptosign"
        # elif msg.authmethod == "anonymous-proxy":
        #     msg.authmethod = "anonymous"
        return super(ProxyBackendSession, self).onWelcome(_msg)

    def onJoin(self, details):
        self.log.info(
            '{func} proxy backend session joined (authmethod={authmethod}, authprovider={authprovider}): '
            'realm="{realm}", authid="{authid}", authrole="{authrole}"',
            func=hltype(self.onJoin),
            realm=hlid(details.realm),
            authid=hlid(details.authid),
            authrole=hlid(details.authrole),
            authmethod=hlval(details.authmethod),
            authprovider=hlval(details.authprovider))

        if not self._on_ready.called:
            self._on_ready.callback(self)

    def onClose(self, wasClean):
        self.log.info('{func} proxy backend session closed (wasClean={wasClean})',
                      func=hltype(self.onClose),
                      wasClean=wasClean)
        if self._frontend is not None and self._frontend.transport is not None:
            try:
                if self._session_id:
                    self._frontend.transport.send(message.Goodbye())
                else:
                    self._frontend.transport.send(message.Abort(ApplicationError.AUTHENTICATION_FAILED))
            except Exception as e:
                self.log.info(
                    "Backend closed, Abort/Goodbye to frontend failed: {fail}",
                    fail=e,
                )
        self._frontend = None
        super(ProxyBackendSession, self).onClose(wasClean)

    def onMessage(self, msg: IMessage):
        if isinstance(msg, (message.Welcome, message.Challenge, message.Abort, message.Goodbye)):
            super(ProxyBackendSession, self).onMessage(msg)
        else:
            # msg is a real WAMP message that our backend WAMP protocol has deserialized. so now we re-serialize it
            # for whatever the frontend is speaking and forward
            self._frontend._forward(msg)


def make_backend_connection(reactor: ReactorBase, controller: 'ProxyController', backend_config: Dict[str, Any],
                            frontend_session: ApplicationSession) -> Deferred:
    """
    Create a connection to a router backend, wiring up the given proxy frontend session
    to forward WAMP in both directions between the frontend and backend sessions.

    Backend configuration example:

        .. code-block:: json
            {
                'auth': {
                    'cryptosign-proxy': {
                        'type': 'static'
                    }
                },
                'transport': {
                    'type': 'rawsocket',
                    'endpoint': {
                        'type': 'tcp',
                        'host': '127.0.0.1',
                        'port': 8442
                    },
                    'serializer': 'cbor',
                    'url': 'rs://localhost'
                }
            }

    :param reactor: Twisted reactor to use.
    :param controller: The proxy controller the backend connection originates from.
    :param cbdir: This node's directory.
    :param backend_config: Proxy backend connection.
    :param frontend_session: The proxy frontend session for which to create a mapped
        backend connection.
    :return: A deferred that resolves with a proxy backend session that is joined on the realm,
        under the authrole, as the proxy frontend session.
    """
    log.info('{func} proxy connecting to backend with backend_config=\n{backend_config}',
             func=hltype(make_backend_connection),
             backend_config=pformat(backend_config))

    cbdir = controller.cbdir

    # fired when the component has connected, authenticated and joined a realm on the backend node
    ready = Deferred()

    # connected node transport
    backend = _create_transport(0, backend_config['transport'])

    # connecting node (this node) private key
    key = _read_node_key(cbdir, private=True)

    # factory for proxy->router backend connections, uses authentication (to router backend worker)
    def create_session():

        # this is our WAMP session to the backend
        session = ProxyBackendSession()

        authextra = {}
        log.debug('{func}::create_session() connecting to backend with authextra=\n{authextra}',
                  func=hltype(make_backend_connection),
                  authextra=pformat(authextra))

        # if auth is configured and includes "cryptosign-proxy", always prefer
        # that and connect to the backend node authenticating with WAMP-cryptosign
        # using the connecting proxy node's key
        #
        # authentication via WAMP-cryptosign SHOULD always be possible with the backend node
        #
        if 'auth' in backend_config and 'cryptosign-proxy' in backend_config['auth']:
            session.add_authenticator(create_authenticator('cryptosign-proxy', privkey=key['hex'],
                                                           authextra=authextra))
            log.debug('{func} using cryptosign-proxy authenticator for backend connection, authextra=\n{authextra}',
                      func=hltype(make_backend_connection),
                      authextra=pformat(authextra))

        # if auth is not configured, or is configured and includes "anonymous-proxy",
        # try to connect to the backend node authenticating with WAMP-anonymous
        #
        # authentication via WAMP-anonymous MAY be possible with the backend node if enabled
        #
        elif 'auth' not in backend_config or 'anonymous-proxy' in backend_config['auth']:
            # IMPORTANT: this is security sensitive! we only allow anonymous proxy
            # locally on a host, that is, when the transport type is Unix domain socket
            if backend_config['transport']['endpoint']['type'] == 'unix':
                session.add_authenticator(create_authenticator('anonymous-proxy', authextra=authextra))
                log.debug(
                    '{func} using anonymous-proxy over UDS authenticator for backend connection, '
                    'authextra=\n{authextra}',
                    func=hltype(make_backend_connection),
                    authextra=pformat(authextra))
            else:
                raise RuntimeError(
                    'anonymous-proxy authenticator only allowed on Unix domain socket based transports, not type "{}"'.
                    format(backend_config['transport']['endpoint']['type']))

        # no valid authentication method found
        else:
            raise RuntimeError('could not determine valid authentication method to connect to the backend node')

        def connected(new_session, transport):
            ready.callback(new_session)

        session.on('connect', connected)
        return session

    # client transport factory to carry our session
    factory = _create_transport_factory(reactor, backend, create_session)
    # reduce noise from logs, otherwise for each connect/disconnect to the backend
    factory.noisy = False
    endpoint = _create_transport_endpoint(reactor, backend_config['transport']['endpoint'])
    transport_d = endpoint.connect(factory)

    def _connected(proto):
        proto._proxy_other_side = frontend_session
        return proto

    def _error(f):
        # backend session disconnected without ever having joined before
        if not ready.called:
            ready.errback(f)

    transport_d.addErrback(_error)
    transport_d.addCallback(_connected)

    return ready


class AuthenticatorSession(ApplicationSession):

    # when running over TLS, require TLS channel binding
    # CHANNEL_BINDING = 'tls-unique'
    CHANNEL_BINDING = None

    def __init__(self, config=None):
        ApplicationSession.__init__(self, config)

        # load the client private key (raw format)
        try:
            self._key = cryptosign.SigningKey.from_key_bytes(config.extra['key'])
        except:
            self.log.failure()
            if self.is_attached():
                self.leave()
        else:
            self.log.info('{func} client public key loaded: {pubkey}',
                          pubkey=hlval(self._key.public_key()),
                          func=hltype(self.__init__))

    def onConnect(self):
        extra = {
            'pubkey': self._key.public_key(),
        }
        self.join(self.config.realm,
                  authmethods=['cryptosign'],
                  authid=self.config.extra.get('authid', None),
                  authextra=extra)

    async def onChallenge(self, challenge):
        try:
            signed_challenge = await self._key.sign_challenge(self, challenge, channel_id_type=self.CHANNEL_BINDING)
            return signed_challenge
        except:
            self.log.failure()
            raise

    def onJoin(self, details):
        if self.config.extra['ready']:
            self.config.extra['ready'].callback(self)
            self.config.extra['ready'] = None

    def onLeave(self, details):
        self.log.info('{func} session closed (details={details})', details=details, func=hltype(self.onDisconnect))

    def onDisconnect(self):
        self.log.info('{func} connection closed', func=hltype(self.onDisconnect))


def make_service_session(reactor: ReactorBase, controller: 'ProxyController', backend_config: Dict[str, Any],
                         realm: str, authrole: str) -> Deferred:
    """
    Create a connection to a router backend, creating a new service session.

    :param reactor: Twisted reactor to use.
    :param controller: The proxy controller this service session is for.
    :param backend_config: Proxy backend connection.
    :param realm: The WAMP realm the service session is joined on.
    :param authrole: The WAMP authrole the service session is joined as.
    :return: A service session joined on the given realm, under the given authrole.
    """
    cbdir = controller.cbdir

    # authid of the proxy session forwarded to the backend: for service session that are
    # not forwarding incoming session (like make_backend_session), but represent an
    # independent session (exposed on the proxy), we synthesize an authid
    proxy_authid = 'proxy-{}'.format(util.generate_serial_number())

    # authid of the connecting backend (proxy service) session is this proxy node's ID
    backend_authid = controller.node_id

    # if auth is configured and includes "cryptosign-proxy", always prefer
    # that and connect to the backend node authenticating with WAMP-cryptosign
    # using the connecting proxy node's key
    #
    # authentication via WAMP-cryptosign SHOULD always be possible with the backend node
    #
    if 'auth' in backend_config and 'cryptosign-proxy' in backend_config['auth']:
        # we will do cryptosign authentication to any backend node

        # FIXME: get node private key from this proxy node
        node_privkey = _read_node_key(cbdir, private=True)['hex']

        authentication = {
            'cryptosign-proxy': {
                'privkey': node_privkey,
                'authid': backend_authid,
                'authextra': {
                    'proxy_realm': realm,
                    'proxy_authid': proxy_authid,
                    'proxy_authrole': authrole,
                    'proxy_authextra': None,
                }
            }
        }

    # if auth is not configured, or is configured and includes "anonymous-proxy",
    # try to connect to the backend node authenticating with WAMP-anonymous
    #
    # authentication via WAMP-anonymous MAY be possible with the backend node if enabled
    #
    elif 'auth' not in backend_config or 'anonymous-proxy' in backend_config['auth']:
        # IMPORTANT: this is security sensitive! we only allow anonymous proxy
        # locally on a host, that is, when the transport type is Unix domain socket
        if backend_config['transport']['endpoint']['type'] == 'unix':
            authentication = {
                'anonymous-proxy': {
                    'authid': backend_authid,
                    'authextra': {
                        'proxy_realm': realm,
                        'proxy_authid': proxy_authid,
                        'proxy_authrole': authrole,
                        'proxy_authextra': None,
                    }
                }
            }
        else:
            raise RuntimeError(
                'anonymous-proxy authenticator only allowed on Unix domain socket based transports, not type "{}"'.
                format(backend_config['transport']['endpoint']['type']))

    # no valid authentication method found
    else:
        raise RuntimeError('could not determine valid authentication method to connect to the backend node')

    # use Component API and create a component for the service session
    comp = Component(transports=[backend_config['transport']], realm=realm, authentication=authentication)

    # fired when the component has connected, authenticated and joined a realm on the backend node
    ready = Deferred()

    @comp.on_join
    def joined(session, details):
        ready.callback(session)

    @comp.on_disconnect
    def disconnect(session, was_clean=False):
        if not ready.called:
            ready.errback(RuntimeError('backend session disconnected without ever having joined before'))

    # start the component and return the component's ready deferred
    comp.start(reactor)
    return ready


STATE_CREATED = 1
STATE_STARTING = 2
STATE_STARTED = 3
STATE_FAILED = 4
STATE_STOPPING = 5
STATE_STOPPED = 6

STATES = {
    STATE_CREATED: "created",
    STATE_STARTING: "starting",
    STATE_STARTED: "started",
    STATE_FAILED: "failed",
    STATE_STOPPING: "stopping",
    STATE_STOPPED: "stopped",
}


class ProxyRoute(object):
    """
    Proxy route run-time representation.
    """
    log = make_logger()

    def __init__(self, controller: 'ProxyController', realm_name: str, route_id: str, config: Dict[str, Any]):
        """

        :param controller: The (proxy) worker controller session the proxy connection is created from.

        :param realm_name: The realm this route applies to.

        :param route_id: The run-time route ID within the proxy worker.

        :param config: The proxy route's configuration, which is a dictionary of role names
            and connection IDs as values.
        """
        self._controller = controller
        self._realm_name = realm_name
        self._route_id = route_id
        self._config = config
        self._started = None
        self._stopped = None
        self._state = STATE_CREATED

    def marshal(self) -> Dict[str, Any]:
        return {
            'realm': self._realm_name,
            'id': self._route_id,
            'config': self._config,
            'started': self._started,
            'stopped': self._stopped,
            'state': self._state,
        }

    def __str__(self):
        return pformat(self.marshal())

    @property
    def realm(self) -> str:
        """

        :return: The realm this route applies to.
        """
        return self._realm_name

    def has_role(self, role_name) -> bool:
        """
        Checks if the given role is mapped in this proxy route.

        :param role_name: Role to lookup.
        :return: ``True`` if the role is configured in this proxy route.
        """
        return role_name in self._config

    def map_connection_id(self, role_name) -> Optional[str]:
        """
        Map the given role to a connection ID according to the configuration of this route.

        :param role_name: Role to map.
        :return: Connection ID configured for the role in this proxy route.
        """
        if role_name in self._config:
            return self._config[role_name]
        else:
            return None

    @property
    def id(self) -> str:
        """

        :return: The ID of this proxy route.
        """
        return self._route_id

    @property
    def config(self) -> Dict[str, Any]:
        """

        :return: The original configuration as supplied to this proxy route.
        """
        return self._config

    @property
    def started(self) -> Optional[int]:
        """

        :return: When this route was started in it's hosting worker.
        """
        return self._started

    @property
    def stopped(self) -> Optional[int]:
        """

        :return: When this route was stopped in it's hosting worker.
        """
        return self._stopped

    @property
    def state(self) -> int:
        """

        :return: Current state of route in it's hosting worker.
        """
        return self._state

    @inlineCallbacks
    def start(self):
        """
        Start proxy route.
        """
        assert self._state == STATE_CREATED
        self._state = STATE_STARTING

        topic = '{}.on_proxy_route_starting'.format(self._controller._uri_prefix)
        yield self._controller.publish(topic, self.marshal(), options=types.PublishOptions(acknowledge=True))

        self._state = STATE_STARTED
        self._started = time_ns()
        self._stopped = None

        topic = '{}.on_proxy_route_started'.format(self._controller._uri_prefix)
        yield self._controller.publish(topic, self.marshal(), options=types.PublishOptions(acknowledge=True))

        self.log.info('{func} proxy route {route_id} started for realm "{realm}" with config=\n{config}',
                      func=hltype(self.start),
                      route_id=hlid(self._route_id),
                      realm=hlval(self._realm_name),
                      config=pformat(self._config))

    @inlineCallbacks
    def stop(self):
        """
        Stop proxy route.
        """
        assert self._state == STATE_STARTED
        self._state = STATE_STOPPING

        topic = '{}.on_proxy_route_stopping'.format(self._controller._uri_prefix)
        yield self._controller.publish(topic, self.marshal(), options=types.PublishOptions(acknowledge=True))

        self._state = STATE_STOPPED
        self._started = None
        self._stopped = time_ns()

        topic = '{}.on_proxy_route_stopped'.format(self._controller._uri_prefix)
        yield self._controller.publish(topic, self.marshal(), options=types.PublishOptions(acknowledge=True))

        self.log.info('{func} proxy route {route_id} stopped for realm "{realm}"',
                      func=hltype(self.stop),
                      route_id=hlid(self._route_id),
                      realm=hlval(self._realm_name))


class ProxyConnection(object):
    """
    Proxy connection run-time representation.
    """
    log = make_logger()

    def __init__(self, controller: 'ProxyController', connection_id: str, config: Dict[str, Any]):
        """
        Example connection configuration for a Unix domain socket based connection using
        WAMP-anonymous proxy authentication:

        .. code-block:: json

            {
                "transport": {
                    "type": "rawsocket",
                    "endpoint": {
                        "type": "unix",
                        "path": "router.sock"
                    },
                    "url": "ws://localhost",
                    "serializer": "cbor"
                },
                "auth": {
                    "anonymous-proxy": {
                        "type": "static"
                    }
                }
            }

        Example connection configuration for a TCP based connection using
        WAMP-cryptosign proxy authentication:

        .. code-block:: json

            {
                "transport": {
                    "type": "rawsocket",
                    "endpoint": {
                        "type": "tcp",
                        "host": "core1",
                        "port": 10023
                    },
                    "url": "ws://core1",
                    "serializer": "cbor"
                },
                "auth": {
                    "cryptosign-proxy": {
                        "type": "static"
                    }
                }
            }

        :param controller: The (proxy) worker controller session the proxy connection is created from.
        :param connection_id: The run-time connection ID within the proxy worker.
        :param config: The proxy connection's configuration.
        """
        self._controller = controller
        self._connection_id = connection_id
        self._config = config
        self._started = None
        self._stopped = None
        self._state = STATE_CREATED

    def marshal(self):
        return {
            'id': self._connection_id,
            'config': self._config,
            'started': self._started,
            'stopped': self._stopped,
            'state': self._state,
        }

    def __str__(self):
        return pformat(self.marshal())

    @property
    def id(self) -> str:
        """

        :return: The ID of this proxy backend connection.
        """
        return self._connection_id

    @property
    def config(self) -> Dict[str, Any]:
        """

        :return: The original configuration as supplied to this proxy backend connection.
        """
        return self._config

    @property
    def started(self) -> Optional[int]:
        """

        :return: When this proxy backend connection was started (Posix time in ns).
        """
        return self._started

    @property
    def stopped(self) -> Optional[int]:
        """

        :return: When this proxy backend connection was stopped (Posix time in ns).
        """
        return self._stopped

    @property
    def state(self) -> int:
        """

        :return: Current state of this proxy backend connection.
        """
        return self._state

    @inlineCallbacks
    def start(self):
        """
        Start this proxy backend connection.
        """
        assert self._state == STATE_CREATED
        self._state = STATE_STARTING

        topic = '{}.on_proxy_connection_starting'.format(self._controller._uri_prefix)
        yield self._controller.publish(topic, self.marshal(), options=types.PublishOptions(acknowledge=True))

        self._state = STATE_STARTED
        self._started = time_ns()
        self._stopped = None

        topic = '{}.on_proxy_connection_started'.format(self._controller._uri_prefix)
        yield self._controller.publish(topic, self.marshal(), options=types.PublishOptions(acknowledge=True))

    @inlineCallbacks
    def stop(self):
        """
        Stop this proxy backend connection.
        """
        assert self._state == STATE_STARTED
        self._state = STATE_STOPPING

        topic = '{}.on_proxy_connection_stopping'.format(self._controller._uri_prefix)
        yield self._controller.publish(topic, self.marshal(), options=types.PublishOptions(acknowledge=True))

        self._state = STATE_STOPPED
        self._started = None
        self._stopped = time_ns()

        topic = '{}.on_proxy_connection_stopped'.format(self._controller._uri_prefix)
        yield self._controller.publish(topic, self.marshal(), options=types.PublishOptions(acknowledge=True))


class ProxyController(TransportController):
    """
    Controller for proxy workers. Manages:

    * **proxy transports**, for accepting incoming client connections
    * **proxy connections**, for backend router connections
    * **proxy routes**, for routes from ``(realm_name, role_name)`` to backend router connections

    and

    * web transport services (from base class `TransportController`), when running a proxy transport
    of type ``web``.

    Proxy controllers also inherit more procedures and events from the base classes

    * :class:`crossbar.worker.transport.TransportController`,
    * :class:`crossbar.worker.controller.WorkerController` and
    * :class:`crossbar.common.process.NativeProcess`.
    """
    WORKER_TYPE = 'proxy'
    WORKER_TITLE = 'WAMP proxy'

    def __init__(self, config=None, reactor=None, personality=None):
        super(ProxyController, self).__init__(
            config=config,
            reactor=reactor,
            personality=personality,
        )
        # the Twisted reactor under which to run
        self._reactor = reactor

        # the node's home directory of this worker
        self._cbdir = config.extra.cbdir

        # map: connection_id -> ProxyConnection
        self._connections: Dict[str, ProxyConnection] = {}

        # map: (realm, authrole) -> Set[ProxyConnection]
        self._connections_by_auth: Dict[Tuple[str, str], Set[ProxyConnection]] = {}

        # map: realm_name -> route_id -> ProxyRoute
        self._routes: Dict[str, Dict[str, ProxyRoute]] = {}

        # map: connection_id -> Set[ProxyRoute]
        self._routes_by_connection: Dict[str, Set[ProxyRoute]] = {}

        # for creating route IDs
        self._next_route_id = 0

        # next route to use in a realm while forwarding connections
        # map: (realm, authrole) -> int
        self._roundrobin_idx = {}

        # since we share some functionality with RouterController we
        # need to have a router_session_factory
        self._router_factory = RouterFactory(
            self.config.extra.node,
            self.config.extra.worker,
            self,  # ProxySession get to ProxyController via .worker here
        )
        self._router_session_factory = RouterSessionFactory(self._router_factory)
        self._router_session_factory.session = ProxyFrontendSession

        # currently mapped session: map of frontend_session => backend_session
        self._backends_by_frontend = {}

        # map: (realm_name, role_name) -> ProxyRoute
        self._service_sessions = {}

    def has_realm(self, realm: str) -> bool:
        """
        Check if a route to a realm with the given name is currently running.

        :param realm: Realm name (the WAMP name, _not_ the run-time object ID).

        :returns: True if a route to the realm (for any role) exists.
        """
        result = realm in self._routes
        self.log.debug('{func}(realm="{realm}") -> {result}',
                       func=hltype(ProxyController.has_realm),
                       realm=hlid(realm),
                       result=hlval(result))
        return result

    def has_role(self, realm: str, authrole: str) -> bool:
        """
        Check if a role with the given name is currently running in the given realm.

        :param realm: WAMP realm (the WAMP name, _not_ the run-time object ID).

        :param authrole: WAMP authentication role (the WAMP URI, _not_ the run-time object ID).

        :returns: True if a route to the realm for the role exists.
        """
        authrole = authrole or 'trusted'
        if realm in self._routes:
            realm_routes = self._routes[realm]

            # the route config is a map with role name as key
            result = any(authrole in route.config for route in realm_routes.values())
        else:
            realm_routes = None
            result = False
        self.log.debug(
            '{func}(realm="{realm}", authrole="{authrole}") -> {result} [routes={routes}, realm_routes={realm_routes}]',
            func=hltype(ProxyController.has_role),
            realm=hlid(realm),
            authrole=hlid(authrole),
            result=hlval(result),
            realm_routes=hlval([route.config for route in realm_routes.values()] if realm_routes else []),
            routes=sorted(self._routes.keys()))
        return result

    @inlineCallbacks
    def get_service_session(self, realm: str, authrole: str) -> ApplicationSession:
        """
        Returns a cached service session on the given realm using the given role.

        Service sessions are used for:

        * access dynamic authenticators (see
            :method:`crossbar.router.auth.pending.PendingAuth._init_dynamic_authenticator`)
        * access the WAMP meta API for the realm
        * forward to/from WAMP for the HTTP bridge

        Service sessions are NOT used to forward WAMP client connections incoming to the proxy worker.

        :param realm: WAMP realm (the WAMP name, _not_ the run-time object ID).

        :param authrole: WAMP authentication role (the WAMP URI, _not_ the run-time object ID).

        :returns: The service session for the realm.
        """
        try:
            self.log.info('{klass}.get_service_session(realm="{realm}", authrole="{authrole}")',
                          klass=self.__class__.__name__,
                          realm=realm,
                          authrole=authrole)

            # create new service session for (realm, authrole) if it doesn't exist yet ..

            # .. check for realm
            if realm not in self._service_sessions:
                if self.has_realm(realm):
                    self.log.info(
                        '{klass}.get_service_session(realm="{realm}", authrole="{authrole}") -> '
                        'not cached, creating new session ..',
                        klass=self.__class__.__name__,
                        realm=realm,
                        authrole=authrole)
                    self._service_sessions[realm] = {}
                else:
                    # mark as non-existing!
                    self._service_sessions[realm] = None

            # .. check for (realm, authrole)
            if self._service_sessions[realm] is not None and authrole not in self._service_sessions[realm]:
                if self.has_role(realm, authrole):
                    # get backend connection configuration selected (round-robin or randomly) from all routes
                    # for the desired (realm, authrole)
                    backend_config = self.get_backend_config(realm, authrole)

                    # create and store a new service session connected to the backend router worker
                    self._service_sessions[realm][authrole] = yield make_service_session(
                        self._reactor, self, backend_config, realm, authrole)
                else:
                    # mark as non-existing!
                    self._service_sessions[realm][authrole] = None

            # return cached service session
            if self._service_sessions[realm] and self._service_sessions[realm][authrole]:
                service_session = self._service_sessions[realm][authrole]
                self.log.info(
                    '{klass}.get_service_session(realm="{realm}", authrole="{authrole}") -> found cached service '
                    'session {session} with authid "{session_authid}" and authrole "{session_authrole}"',
                    klass=self.__class__.__name__,
                    realm=realm,
                    authrole=authrole,
                    session=service_session.session_id,
                    session_authid=service_session.authid,
                    session_authrole=service_session.authrole)
                return service_session
            else:
                self.log.warn(
                    '{klass}.get_service_session(realm="{realm}", authrole="{authrole}") -> no such realm/authrole!',
                    klass=self.__class__.__name__,
                    realm=realm,
                    authrole=authrole)
                return None
        except:
            self.log.failure()
            raise

    def can_map_backend(self, session_id, realm, authid, authrole, authextra):
        """
        Checks if the proxy can map the incoming frontend session to a backend.

        :returns: True only-if map_backend() can succeed later for the
            same args (essentially, if the realm + role exist).
        """
        return self.has_realm(realm) and self.has_role(realm, authrole)

    @inlineCallbacks
    def map_backend(self, frontend, realm: str, authid: str, authrole: str, authextra: Optional[Dict[str, Any]]):
        """
        Returns the cached backend forwarding session for the given frontend session.
        Map the given frontend session to a backend session under the given
        authentication credentials.

        :param frontend:
        :param realm:
        :param authid:
        :param authrole:
        :param authextra:
        :return: a protocol instance connected to the backend
        """
        self.log.debug(
            '{func}(frontend={frontend}, realm="{realm}", authid="{authid}", authrole="{authrole}", '
            'authextra={authextra})',
            func=hltype(self.map_backend),
            frontend=frontend,
            realm=hlid(realm),
            authid=hlid(authid),
            authrole=hlid(authrole),
            authextra=authextra)

        if frontend in self._backends_by_frontend:
            backend = self._backends_by_frontend[frontend]
            self.log.info('{func} CACHE HIT {backend}', func=hltype(self.map_backend), backend=backend)
            return backend

        backend_config = self.get_backend_config(realm, authrole)

        # if auth uses cryptosign but has no privkey, we'd ideally
        # insert the node's private key

        if authrole is None:
            if len(self._routes.get(realm, set())) != 1:
                raise RuntimeError("Cannot select default role unless realm has exactly 1")

        self.log.debug(
            '{func} CACHE MISS - opening new proxy backend connection for realm "{realm}", authrole "{authrole}" '
            'using backend_config=\n{backend_config}',
            func=hltype(self.map_backend),
            backend_config=pformat(backend_config),
            realm=hlid(realm),
            authrole=hlid(authrole))

        try:
            backend_proto = yield make_backend_connection(self._reactor, self, backend_config, frontend)
        except DNSLookupError as e:
            self.log.warn('{func} proxy worker could not connect to router backend: DNS resolution failed ({error})',
                          func=hltype(self.map_backend),
                          error=str(e))
            raise e

        if frontend:
            self._backends_by_frontend[frontend] = backend_proto

        self.log.debug(
            '{func} proxy backend connection opened mapping frontend session to realm "{realm}", authrole "{authrole}"',
            func=hltype(self.map_backend),
            backend_config=pformat(backend_config),
            realm=hlid(realm),
            authrole=hlid(authrole))

        returnValue(backend_proto)

    def unmap_backend(self, frontend, backend):
        """
        Unmap the backend session from the given frontend session it is currently mapped to.
        """
        self.log.debug('{func}(frontend={frontend}, backend={backend})',
                       func=hltype(self.unmap_backend),
                       frontend=frontend,
                       backend=backend)
        if frontend in self._backends_by_frontend:
            if self._backends_by_frontend[frontend] == backend:
                # alright, the given frontend is indeed currently mapped to the given backend session: close the
                # session and delete it
                backend.leave()
                del self._backends_by_frontend[frontend]
                self.log.debug(
                    '{func} unmapped frontend session {frontend_session_id} from backend session {backend_session_id}',
                    func=hltype(self.unmap_backend),
                    frontend_session_id=hlid(frontend._session_id),
                    backend_session_id=hlid(backend._session_id))
            else:
                self.log.warn('{func} frontend session {frontend_session_id} currently mapped to backend session '
                              '{backend_session_id} - NOT to specified backend {specified_session_id}'.format(
                                  func=hltype(self.unmap_backend),
                                  frontend_session_id=hlid(frontend._session_id),
                                  backend_session_id=hlid(self._backends_by_frontend[frontend]._session_id),
                                  specified_session_id=hlid(backend._session_id)))
        else:
            if frontend:
                self.log.warn('{func} frontend session {session_id} not currently mapped to any backend',
                              func=hltype(self.unmap_backend),
                              session_id=hlid(frontend._session_id))

    def get_backend_config(self, realm_name, role_name):
        """
        Return backend connection information for the given backend realm and role.

        :returns: a dict containing the connection configuration for the backend
            identified by the realm_name and role_name
        """
        assert self.has_role(realm_name, role_name)

        key = realm_name, role_name
        if key not in self._roundrobin_idx:
            self._roundrobin_idx[key] = 0
        else:
            self._roundrobin_idx[key] += 1

        idx = self._roundrobin_idx[key] % len(self._connections_by_auth[key])
        connection = list(self._connections_by_auth[key])[idx]

        return connection.config

    @inlineCallbacks
    def onJoin(self, details):
        """
        Called when worker process has joined the node's management realm.
        """
        self.log.info(
            '{func} proxy worker "{worker_id}" session {session_id} initializing',
            func=hltype(self.onJoin),
            worker_id=hlid(self._worker_id),
            session_id=hlid(details.session),
        )

        yield WorkerController.onJoin(self, details, publish_ready=False)

        yield self.publish_ready()

    @wamp.register(None)
    def get_proxy_transports(self, details=None):
        """
        Get proxy (listening) transports currently running in this proxy worker.

        :param details: Call details.
        :type details: :class:`autobahn.wamp.types.CallDetails`

        :returns: List of transport IDs of transports currently running.
        :rtype: list
        """
        self.log.debug('{func}(caller_authid="{caller_authid}")',
                       func=hltype(self.get_proxy_transports),
                       caller_authid=hlval(details.caller_authid))
        return sorted(self.transports.keys())

    @wamp.register(None)
    def get_proxy_transport(self, transport_id, details=None):
        """
        Get transport currently running in this proxy worker.

        :param details: Call details.
        :type details: :class:`autobahn.wamp.types.CallDetails`

        :returns: List of transports currently running.
        :rtype: dict
        """
        self.log.debug('{func}(transport_id={transport_id})',
                       func=hltype(self.get_proxy_transport),
                       transport_id=hlid(transport_id),
                       caller_authid=hlval(details.caller_authid))

        if transport_id in self.transports:
            transport = self.transports[transport_id]
            return transport.marshal()
        else:
            raise ApplicationError("crossbar.error.no_such_object", "No transport {}".format(transport_id))

    @inlineCallbacks
    @wamp.register(None)
    def start_proxy_transport(self, transport_id, config, details=None):
        """
        Start a new proxy front-end listening transport.

        :param transport_id: The run-time ID to start the transport under.
        :param config: The listening transport configuration.
        :param details: WAMP call details.

        :return: Proxy transport run-time metadata.
        """
        self.log.info('{func}(transport_id="{transport_id}", config={config})',
                      func=hltype(self.start_proxy_transport),
                      transport_id=hlid(transport_id),
                      config='...',
                      caller_authid=hlval(details.caller_authid))

        # prohibit starting a transport twice
        if transport_id in self.transports:
            _emsg = 'Could not start transport: a transport with ID "{}" is already running (or starting)'.format(
                transport_id)
            self.log.error(_emsg)
            raise ApplicationError('crossbar.error.already_running', _emsg)

        # create a transport and parse the transport configuration
        # (NOTE: yes, this is re-using create_router_transport so we
        # can proxy every service a 'real' router can)
        proxy_transport = self.personality.create_router_transport(self, transport_id, config)

        caller = details.caller if details else None
        transport_started = proxy_transport.marshal()
        self.publish('{}.on_proxy_transport_starting'.format(self._uri_prefix),
                     transport_started,
                     options=types.PublishOptions(exclude=caller))

        # start listening ..
        try:
            yield proxy_transport.start(False)
        except Exception as err:
            _emsg = "Cannot listen on transport endpoint: {log_failure}"
            self.log.error(_emsg, log_failure=err)

            self.publish('{}.on_proxy_transport_stopped'.format(self._uri_prefix),
                         transport_started,
                         options=types.PublishOptions(exclude=caller))

            raise ApplicationError("crossbar.error.cannot_listen", _emsg.format(log_failure=err))

        self.transports[transport_id] = proxy_transport

        self.publish('{}.on_proxy_transport_started'.format(self._uri_prefix),
                     transport_started,
                     options=types.PublishOptions(exclude=caller))

        self.log.info('{func} proxy transport "{transport_id}" started and listening!',
                      func=hltype(self.start_proxy_transport),
                      transport_id=hlid(transport_id))

        returnValue(proxy_transport.marshal())

    @inlineCallbacks
    @wamp.register(None)
    def stop_proxy_transport(self, transport_id, details=None):
        """
        Stop a currently running proxy front-end listening transport.

        :param transport_id: The run-time ID of the transport to stop.
        :param details: WAMP call details.
        :return: Proxy transport run-time information.
        """
        if transport_id not in self._transports:
            raise ApplicationError('crossbar.error.no_such_object',
                                   'no proxy transport with ID "{}" currently running'.format(transport_id))

        caller = details.caller if details else None
        transport_stopped = self._transports[transport_id].marshal()

        self.publish('{}.on_proxy_transport_stopping'.format(self._uri_prefix),
                     transport_stopped,
                     options=types.PublishOptions(exclude=caller))

        yield self._transports[transport_id].port.stopListening()
        del self._transports[transport_id]

        self.publish('{}.on_proxy_transport_stopping'.format(self._uri_prefix),
                     transport_stopped,
                     options=types.PublishOptions(exclude=caller))

        return transport_stopped

    @wamp.register(None)
    def get_proxy_routes(self, details=None):
        """
        Get proxy routes currently running in this proxy worker.

        :param details: Call details.
        :type details: :class:`autobahn.wamp.types.CallDetails`

        :returns: List of (target) realm names in proxy routes currently running.
        :rtype: list
        """
        self.log.debug('{func}(caller_authid="{caller_authid}")',
                       func=hltype(self.get_proxy_routes),
                       caller_authid=hlval(details.caller_authid))
        return sorted(self._routes.keys())

    @wamp.register(None)
    def list_proxy_realm_routes(self, realm_name, details=None):
        """
        Get list of all routes enabled for a particular realm
        """
        if realm_name in self._routes:
            return [self._routes[realm_name][route_id].marshal() for route_id in self._routes[realm_name].keys()]
        else:
            raise ApplicationError("crossbar.error.no_such_object",
                                   'No route for realm "{}" in proxy'.format(realm_name))

    @wamp.register(None)
    def get_proxy_realm_route(self, realm_name, route_id, details=None):
        """
        Get a particular realm-route

        :param details: Call details.
        :type details: :class:`autobahn.wamp.types.CallDetails`

        :returns: Proxy route object.
        :rtype: dict
        """
        self.log.debug('{func}(realm_name={realm_name})',
                       func=hltype(self.get_proxy_realm_route),
                       realm_name=hlid(realm_name),
                       caller_authid=hlval(details.caller_authid))

        try:
            return self._routes[realm_name][route_id]
        except KeyError:
            raise ApplicationError("crossbar.error.no_such_object",
                                   'No route "{}" for realm "{}" in proxy'.format(route_id, realm_name))

    @inlineCallbacks
    @wamp.register(None)
    def start_proxy_realm_route(self, realm_name, config, details=None):
        """
        Start a new proxy route for the given realm. A proxy route maps authroles
        on the given realm to proxy backend connection IDs.

        Example route configuration:

        .. code-block:: json

            {
                "anonymous": "conn1",
                "restbridge": "conn1",
                "user": "conn2"
            }

        In this example, the two specified connections ``"conn1"`` and ``"conn2"``
        must be running already.

        :param realm_name: The realm this route should apply for.
        :param config: The route configuration.
        :param details: WAMP call details.
        :return: Proxy route run-time information.
        """
        self.log.info(
            '{func}(realm_name="{realm_name}", config={config})',
            func=hltype(self.start_proxy_realm_route),
            realm_name=realm_name,
            config=config,
        )

        # check that we already know about all connections specified in the route
        connection_ids = set()
        for role_name in config.keys():
            connection_id = config[role_name]
            if connection_id not in self._connections:
                raise ApplicationError(
                    "crossbar.error.no_such_object",
                    'no connection "{}" found for realm "{}" and role "{}" in proxy route config'.format(
                        connection_id, realm_name, role_name))
            else:
                connection_ids.add(connection_id)

        # remember connections mapped from proxy routes by (realm, authrole)
        for role_name in config.keys():
            connection_id = config[role_name]
            connection = self._connections[connection_id]
            key = (realm_name, role_name)
            if key not in self._connections_by_auth:
                self._connections_by_auth[key] = set()
            self._connections_by_auth[key].add(connection)

        if realm_name not in self._routes:
            self._routes[realm_name] = dict()

        route_id = 'route{:03d}'.format(self._next_route_id)
        self._next_route_id += 1

        route = ProxyRoute(self, realm_name, route_id, config)
        yield route.start()

        # remember route by route ID
        self._routes[realm_name][route_id] = route

        # remember route by connections
        for connection_id in connection_ids:
            if connection_id not in self._routes_by_connection:
                self._routes_by_connection[connection_id] = set()
            self._routes_by_connection[connection_id].add(route)

        returnValue(route.marshal())

    @inlineCallbacks
    @wamp.register(None)
    def stop_proxy_realm_route(self, realm_name, route_id, details=None):
        """
        Stop a currently running proxy route.

        :param realm_name: The name of the realm to stop the route for.
        :param route_id: Which route to stop
        :param details: WAMP call details.
        :return: Run-time information about the stopped route.
        """
        self.log.info('{func}(realm_name={realm_name}, caller_authid="{caller_authid}")',
                      func=hltype(self.stop_proxy_realm_route),
                      realm_name=realm_name,
                      caller_authid=hlval(details.caller_authid))
        if realm_name not in self._routes:
            raise ApplicationError('crossbar.error.no_such_object',
                                   'no proxy routes for realm "{}" currently running'.format(realm_name))
        if route_id not in self._routes[realm_name]:
            raise ApplicationError('crossbar.error.no_such_object',
                                   'no route "{}" for realm "{}" currently running'.format(route_id, realm_name))

        route = self._routes[realm_name][route_id]
        yield route.stop()
        del self._routes[realm_name][route_id]

        # If all routes are stopped, clear the realm from routes map
        # Relevant discussion: https://github.com/crossbario/crossbar/pull/1968
        if len(self._routes[realm_name]) == 0:
            del self._routes[realm_name]

        returnValue(route.marshal())

    @wamp.register(None)
    def get_proxy_connections(self, details=None):
        """
        Get currently running proxy connections.

        :param details: WAMP call details.
        :return: List of run-time IDs of currently running connection.s
        """
        self.log.debug('{func}(caller_authid="{caller_authid}")',
                       func=hltype(self.get_proxy_connections),
                       caller_authid=hlval(details.caller_authid))

        return sorted(self._connections.keys())

    @wamp.register(None)
    def get_proxy_connection(self, connection_id, details=None):
        """
        Get run-time information for a currently running proxy connection.

        :param connection_id: The run-time ID of the proxy connection to return information for.
        :param details: WAMP call details.
        :return: Proxy connection configuration.
        """
        self.log.debug('{func}(connection_id={connection_id}, caller_authid="{caller_authid}")',
                       func=hltype(self.get_proxy_connection),
                       connection_id=hlid(connection_id),
                       caller_authid=hlval(details.caller_authid))

        if connection_id in self._connections:
            connection = self._connections[connection_id]
            return connection.marshal()
        else:
            raise ApplicationError("crossbar.error.no_such_object",
                                   'no proxy connection with ID "{}" currently running'.format(connection_id))

    @inlineCallbacks
    @wamp.register(None)
    def start_proxy_connection(self, connection_id, config, details=None):
        """
        Start a new backend connection for the proxy.

        Called from master node orchestration in
            :method:`crossbar.master.arealm.arealm.ApplicationRealmMonitor._apply_webcluster_connections`.

        :param connection_id:
        :param config:
        :param details:
        :return:
        """
        self.log.info('{func}(connection_id={connection_id}, config=.., caller_authid={caller_authid}):\n{config}',
                      func=hltype(self.start_proxy_connection),
                      connection_id=connection_id,
                      config=pformat(config),
                      caller_authid=hlval(details.caller_authid))
        if connection_id in self._connections:
            raise ApplicationError('crossbar.error.already_running',
                                   'proxy connection with ID "{}" already running'.format(connection_id))

        connection = ProxyConnection(self, connection_id, config)
        self._connections[connection_id] = connection
        yield connection.start()

        returnValue(connection.marshal())

    @inlineCallbacks
    @wamp.register(None)
    def stop_proxy_connection(self, connection_id, details=None):
        """

        :param connection_id:
        :param details:
        :return:
        """
        self.log.info('{func}(connection_id={connection_id}, caller_authid="{caller_authid}")',
                      func=hltype(self.stop_proxy_connection),
                      connection_id=connection_id,
                      caller_authid=hlval(details.caller_authid))
        if connection_id not in self._connections:
            raise ApplicationError('crossbar.error.no_such_object',
                                   'no proxy connection with ID "{}" currently running'.format(connection_id))

        connection = self._connections[connection_id]
        yield connection.stop()
        del self._connections[connection_id]

        returnValue(connection.marshal())


IRealmContainer.register(ProxyController)



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################

import os
import datetime
import json
import hmac
import hashlib
import base64
import binascii
from typing import Dict, Any

from autobahn.wamp.exception import ApplicationError
from autobahn.twisted.wamp import ApplicationSession

from txaio import make_logger

from crossbar._util import dump_json
from crossbar._compat import native_string
from crossbar._log_categories import log_categories
from ipaddress import ip_address, ip_network

from twisted.web import server
from twisted.web.resource import Resource

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import hmac as hazmat_hmac

from autobahn.websocket.utf8validator import Utf8Validator

_validator = Utf8Validator()

_ALLOWED_CONTENT_TYPES = set([b'application/json'])


class _InvalidUnicode(BaseException):
    """
    Invalid Unicode was found.
    """


# used for constant-time compares
_nonce = os.urandom(32)


def _hmac_sha256(key, data):
    """
    :returns: the HMAC-SHA256 of 'data' using 'key'
    """
    h = hazmat_hmac.HMAC(key, hashes.SHA256(), default_backend())
    h.update(data)
    return h.finalize()


def _constant_compare(a, b):
    """
    Compare the two byte-strings 'a' and 'b' using a constant-time
    algorithm. The byte-strings should be the same length.
    """
    return _hmac_sha256(_nonce, a.encode('ascii')) == _hmac_sha256(_nonce, b.encode('ascii'))


def _confirm_github_signature(request, secret_token, raw_body):
    """
    confirm that signature headers from GitHub are present and valid
    """
    # stuff just "won't work" unless we gives bytes objects to the
    # underlying crypto primitives
    if not isinstance(secret_token, bytes):
        secret_token = secret_token.encode('ascii')
    assert isinstance(raw_body, bytes)
    # must have the header to continue
    if not request.requestHeaders.getRawHeaders('X-Hub-Signature'):
        return False
    purported_signature = str(request.requestHeaders.getRawHeaders('X-Hub-Signature')[0]).lower()
    # NOTE: never use SHA1 for new code ... but GitHub signatures are
    # SHA1, so we have to here :(
    h = hazmat_hmac.HMAC(secret_token, hashes.SHA1(), default_backend())  # nosec
    h.update(raw_body)
    our_signature = "sha1={}".format(binascii.b2a_hex(h.finalize()).decode('ascii'))
    return _constant_compare(our_signature, purported_signature)


class _CommonResource(Resource):
    """
    Shared components between PublisherResource and CallerResource.
    """
    isLeaf = True
    decode_as_json = True

    def __init__(self, options: Dict[str, Any], session: ApplicationSession):
        """

        :param options: Options for path service from configuration.
        :param session: WAMP session to be used for forwarding events / calls.
        """
        Resource.__init__(self)
        self._options = options
        self._session = session
        self._debug = False
        self.log = make_logger()

        if 'debug' in options and options['debug']:
            self._debug = True

        self._key = None
        if 'key' in options:
            self._key = options['key'].encode('utf8')

        self._secret = None
        if 'secret' in options:
            self._secret = options['secret'].encode('utf8')

        self._post_body_limit = int(options.get('post_body_limit', 0))
        self._timestamp_delta_limit = int(options.get('timestamp_delta_limit', 300))

        self._require_ip = None
        if 'require_ip' in options:
            self._require_ip = [ip_network(net) for net in options['require_ip']]

        self._require_tls = options.get('require_tls', None)

    def _deny_request(self, request, code, **kwargs):
        """
        Called when client request is denied.
        """
        if "log_category" not in kwargs.keys():
            kwargs["log_category"] = "AR" + str(code)

        self.log.debug(code=code, **kwargs)

        error_str = log_categories[kwargs['log_category']].format(**kwargs)
        body = dump_json({"error": error_str, "args": [], "kwargs": {}}, True).encode('utf8')
        request.setResponseCode(code)
        return body

    def _fail_request(self, request, **kwargs):
        """
        Called when client request fails.
        """
        res = {}
        err = kwargs["failure"]
        if isinstance(err.value, ApplicationError):
            res['error'] = err.value.error
            if err.value.args:
                res['args'] = err.value.args
            else:
                res['args'] = []
            if err.value.kwargs:
                res['kwargs'] = err.value.kwargs
            else:
                res['kwargs'] = {}

            # This is a user-level error, not a CB error, so return 200
            code = 200
        else:
            # This is a "CB" error, so return 500 and a generic error
            res['error'] = 'wamp.error.runtime_error'
            res['args'] = ["Sorry, Crossbar.io has encountered a problem."]
            res['kwargs'] = {}

            # CB-level error, return 500
            code = 500

            self.log.failure(None, failure=err, log_category="AR500")

        body = json.dumps(res).encode('utf8')

        if "log_category" not in kwargs.keys():
            kwargs["log_category"] = "AR" + str(code)

        self.log.debug(code=code, **kwargs)

        request.setResponseCode(code)
        request.write(body)
        request.finish()

    def _complete_request(self, request, code, body, **kwargs):
        """
        Called when client request is complete.
        """
        if "log_category" not in kwargs.keys():
            kwargs["log_category"] = "AR" + str(code)

        self.log.debug(code=code, **kwargs)
        request.setResponseCode(code)
        request.write(body)

    def _set_common_headers(self, request):
        """
        Set common HTTP response headers.
        """
        origin = request.getHeader(b'origin')
        if origin is None or origin == b'null':
            origin = b'*'

        request.setHeader(b'access-control-allow-origin', origin)
        request.setHeader(b'access-control-allow-credentials', b'true')
        request.setHeader(b'cache-control', b'no-store,no-cache,must-revalidate,max-age=0')
        request.setHeader(b'content-type', b'application/json; charset=UTF-8')

        headers = request.getHeader(b'access-control-request-headers')
        if headers is not None:
            request.setHeader(b'access-control-allow-headers', headers)

    def render(self, request):
        """
        Handle the request. All requests start here.
        """
        self.log.debug(log_category="AR100", method=request.method, path=request.path)
        self._set_common_headers(request)

        try:
            if request.method not in (b"POST", b"PUT", b"OPTIONS"):
                return self._deny_request(request, 405, method=request.method, allowed="POST, PUT")
            else:

                if request.method == b"OPTIONS":
                    # http://greenbytes.de/tech/webdav/rfc2616.html#rfc.section.14.7
                    request.setHeader(b'allow', b'POST,PUT,OPTIONS')

                    # https://www.w3.org/TR/cors/#access-control-allow-methods-response-header
                    request.setHeader(b'access-control-allow-methods', b'POST,PUT,OPTIONS')

                    request.setResponseCode(200)
                    return b''
                else:
                    return self._render_request(request)
        except Exception as e:
            self.log.failure(log_category="CB501", exc=e)
            return self._deny_request(request, 500, log_category="CB500")

    def _render_request(self, request):
        """
        Receives an HTTP/POST|PUT request, and then calls the Publisher/Caller
        processor.
        """
        # read HTTP/POST|PUT body
        body = request.content.read()

        args = {native_string(x): y[0] for x, y in request.args.items()}
        headers = request.requestHeaders

        # check content type + charset encoding
        #
        content_type_header = headers.getRawHeaders(b"content-type", [])

        if len(content_type_header) > 0:
            content_type_elements = [x.strip().lower() for x in content_type_header[0].split(b";")]
        else:
            content_type_elements = []

        if self.decode_as_json:
            # if the client sent a content type, it MUST be one of _ALLOWED_CONTENT_TYPES
            # (but we allow missing content type .. will catch later during JSON
            # parsing anyway)
            if len(content_type_elements) > 0 and \
               content_type_elements[0] not in _ALLOWED_CONTENT_TYPES:
                return self._deny_request(request,
                                          400,
                                          accepted=list(_ALLOWED_CONTENT_TYPES),
                                          given=content_type_elements[0],
                                          log_category="AR452")

        encoding_parts = {}

        if len(content_type_elements) > 1:
            try:
                for item in content_type_elements:
                    if b"=" not in item:
                        # Don't bother looking at things "like application/json"
                        continue

                    # Parsing things like:
                    # charset=utf-8
                    _ = native_string(item).split("=")
                    assert len(_) == 2

                    # We don't want duplicates
                    key = _[0].strip().lower()
                    assert key not in encoding_parts
                    encoding_parts[key] = _[1].strip().lower()
            except:
                return self._deny_request(request, 400, log_category="AR450")

        charset_encoding = encoding_parts.get("charset", "utf-8")

        if charset_encoding not in ["utf-8", 'utf8']:
            return self._deny_request(request, 400, log_category="AR450")

        # enforce "post_body_limit"
        #
        body_length = len(body)
        content_length_header = headers.getRawHeaders(b"content-length", [])

        if len(content_length_header) == 1:
            content_length = int(content_length_header[0])
        elif len(content_length_header) > 1:
            return self._deny_request(request, 400, log_category="AR463")
        else:
            content_length = body_length

        if body_length != content_length:
            # Prevent the body length from being different to the given
            # Content-Length. This is so that clients can't lie and bypass
            # length restrictions by giving an incorrect header with a large
            # body.
            return self._deny_request(request, 400, bodylen=body_length, conlen=content_length, log_category="AR465")

        if self._post_body_limit and content_length > self._post_body_limit:
            return self._deny_request(request, 413, length=content_length, accepted=self._post_body_limit)

        #
        # if we were given a GitHub token, check for a valid signature
        # header
        #

        github_secret = self._options.get("github_secret", "")
        if github_secret:
            if not _confirm_github_signature(request, github_secret, body):
                return self._deny_request(
                    request,
                    400,
                    log_cagegory="AR467",
                )

        #
        # parse/check HTTP/POST|PUT query parameters
        #

        # key
        #
        if 'key' in args:
            key_str = args["key"]
        else:
            if self._secret:
                return self._deny_request(request, 400, reason="'key' field missing", log_category="AR461")

        # timestamp
        #
        if 'timestamp' in args:
            timestamp_str = args["timestamp"]
            try:
                ts = datetime.datetime.strptime(native_string(timestamp_str), "%Y-%m-%dT%H:%M:%S.%fZ")
                delta = abs((ts - datetime.datetime.utcnow()).total_seconds())
                if self._timestamp_delta_limit and delta > self._timestamp_delta_limit:
                    return self._deny_request(request, 400, log_category="AR464")
            except ValueError:
                return self._deny_request(
                    request,
                    400,
                    reason="invalid timestamp '{0}' (must be UTC/ISO-8601, e.g. '2011-10-14T16:59:51.123Z')".format(
                        native_string(timestamp_str)),
                    log_category="AR462")
        else:
            if self._secret:
                return self._deny_request(request,
                                          400,
                                          reason="signed request required, but mandatory 'timestamp' field missing",
                                          log_category="AR461")

        # seq
        #
        if 'seq' in args:
            seq_str = args["seq"]
            try:
                # FIXME: check sequence
                seq = int(seq_str)  # noqa
            except:
                return self._deny_request(request,
                                          400,
                                          reason="invalid sequence number '{0}' (must be an integer)".format(
                                              native_string(seq_str)),
                                          log_category="AR462")
        else:
            if self._secret:
                return self._deny_request(request, 400, reason="'seq' field missing", log_category="AR461")

        # nonce
        #
        if 'nonce' in args:
            nonce_str = args["nonce"]
            try:
                # FIXME: check nonce
                nonce = int(nonce_str)  # noqa
            except:
                return self._deny_request(request,
                                          400,
                                          reason="invalid nonce '{0}' (must be an integer)".format(
                                              native_string(nonce_str)),
                                          log_category="AR462")
        else:
            if self._secret:
                return self._deny_request(request, 400, reason="'nonce' field missing", log_category="AR461")

        # signature
        #
        if 'signature' in args:
            signature_str = args["signature"]
        else:
            if self._secret:
                return self._deny_request(request, 400, reason="'signature' field missing", log_category="AR461")

        # do more checks if signed requests are required
        #
        if self._secret:

            if key_str != self._key:
                return self._deny_request(request,
                                          401,
                                          reason="unknown key '{0}' in signed request".format(native_string(key_str)),
                                          log_category="AR460")

            # Compute signature: HMAC[SHA256]_{secret} (key | timestamp | seq | nonce | body) => signature
            hm = hmac.new(self._secret, None, hashlib.sha256)
            hm.update(key_str)
            hm.update(timestamp_str)
            hm.update(seq_str)
            hm.update(nonce_str)
            hm.update(body)
            signature_recomputed = base64.urlsafe_b64encode(hm.digest())

            if signature_str != signature_recomputed:
                return self._deny_request(request, 401, log_category="AR459")
            else:
                self.log.debug("REST request signature valid.", log_category="AR203")

        # user_agent = headers.get("user-agent", "unknown")
        client_ip = request.getClientIP()
        is_secure = request.isSecure()

        # enforce client IP address
        #
        if self._require_ip:
            ip = ip_address(client_ip)
            allowed = False
            for net in self._require_ip:
                if ip in net:
                    allowed = True
                    break
            if not allowed:
                return self._deny_request(request, 400, log_category="AR466")

        # enforce TLS
        #
        if self._require_tls:
            if not is_secure:
                return self._deny_request(request, 400, reason="request denied because not using TLS")

        # FIXME: authorize request
        authorized = True

        if not authorized:
            return self._deny_request(request, 401, reason="not authorized")

        _validator.reset()
        validation_result = _validator.validate(body)

        # validate() returns a 4-tuple, of which item 0 is whether it
        # is valid
        if not validation_result[0]:
            return self._deny_request(request, 400, log_category="AR451")

        event = body.decode('utf8')

        if self.decode_as_json:
            try:
                event = json.loads(event)
            except Exception as e:
                return self._deny_request(request, 400, exc=e, log_category="AR453")

            if not isinstance(event, dict):
                return self._deny_request(request, 400, log_category="AR454")

        d = self._process(request, event)

        if isinstance(d, bytes):
            # If it's bytes, return it directly
            return d
        else:
            # If it's a Deferred, let it run.
            d.addCallback(lambda _: request.finish())

        return server.NOT_DONE_YET

    def _process(self, request, event):
        raise NotImplementedError()



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################

from typing import Dict, Any, Union

from twisted.internet.defer import inlineCallbacks

from autobahn.wamp.types import ComponentConfig
from autobahn.twisted.wamp import ApplicationSession
from autobahn.wamp.exception import ApplicationError

from crossbar.bridge.rest import PublisherResource, CallerResource
from crossbar.bridge.rest import WebhookResource

from crossbar.webservice.base import RouterWebService

from crossbar.worker.proxy import ProxyController
from crossbar.worker.router import RouterController

__all__ = ('RouterWebServiceRestPublisher', 'RouterWebServiceRestCaller', 'RouterWebServiceWebhook')


@inlineCallbacks
def _create_resource(resource_klass: Union[PublisherResource, CallerResource, WebhookResource], worker,
                     config: Dict[str, Any]):
    """
    Create a new Twisted Web resource wrapping a WAMP action.

    :param resource_klass: WAMP action wrapper class.
    :param transport: Web transport
    :param config:
    :return: Twisted Web resource
    """
    # the WAMP realm and authid to use for the WAMP session backing this resource
    realm = config['realm']
    authrole = config.get('role', 'anonymous')

    # create a service session on the router/proxy controller: this will be used
    # to forward HTTP requests to WAMP
    if isinstance(worker, RouterController):
        session = ApplicationSession(ComponentConfig(realm=realm, extra=None))
        router = worker._router_session_factory._routerFactory._routers[realm]
        worker._router_session_factory.add(session, router, authrole=authrole)
    elif isinstance(worker, ProxyController):
        if not worker.has_realm(realm):
            raise ApplicationError('crossbar.error.no_such_object',
                                   'no realm "{}" in configured routes of proxy worker'.format(realm))
        if not worker.has_role(realm, authrole):
            raise ApplicationError(
                'crossbar.error.no_such_object',
                'no role "{}" on realm "{}" in configured routes of proxy worker'.format(authrole, realm))
        session = yield worker.get_service_session(realm, authrole)
        if not session or not session.is_attached():
            raise ApplicationError(
                'crossbar.error.cannot_start',
                'could not attach service session for HTTP bridge (role "{}" on realm "{}")'.format(authrole, realm))

        # session.authextra:
        # {'x_cb_node': 'intel-nuci7-61704', 'x_cb_worker': 'worker001', 'x_cb_peer': 'unix:None', 'x_cb_pid': 61714}

        assert session.realm == realm, 'service session: requested realm "{}", but got "{}"'.format(
            realm, session.realm)
        assert session.authrole == authrole, 'service session: requested authrole "{}", but got "{}"'.format(
            authrole, session.authrole)
    else:
        assert False, 'logic error: unexpected worker type {} in RouterWebServiceRestCaller.create'.format(
            type(worker))

    # now create the caller Twisted Web resource
    resource = resource_klass(config.get('options', {}), session)

    return resource


class RouterWebServiceRestPublisher(RouterWebService):
    """
    HTTP/REST-to-WAMP Publisher Web service (part of REST-bridge).
    """
    @staticmethod
    @inlineCallbacks
    def create(transport, path: str, config: Dict[str, Any]):
        """
        Create a new HTTP/REST-to-WAMP Publisher Web service (part of REST-bridge).

        :param transport: Web transport on which to add the web service.
        :param path: HTTP path on which to add the web service.
        :param config: Web service configuration.
        :return: Web service instance.
        """
        personality = transport.worker.personality
        personality.WEB_SERVICE_CHECKERS['publisher'](personality, config)

        # the realm container on which to add the resource
        worker: Union[RouterController, ProxyController] = transport._worker

        resource = yield _create_resource(PublisherResource, worker, config)
        return RouterWebServiceRestPublisher(transport, path, config, resource)


class RouterWebServiceRestCaller(RouterWebService):
    """
    HTTP/REST-to-WAMP Caller Web service (part of REST-bridge).
    """
    @staticmethod
    @inlineCallbacks
    def create(transport, path: str, config: Dict[str, Any]):
        """
        Create a new HTTP/REST-to-WAMP Caller Web service (part of REST-bridge).

        :param transport: Web transport on which to add the web service.
        :param path: HTTP path on which to add the web service.
        :param config: Web service configuration.
        :return: Web service instance.
        """
        personality = transport.worker.personality
        personality.WEB_SERVICE_CHECKERS['caller'](personality, config)

        # the realm container on which to add the resource
        worker: Union[RouterController, ProxyController] = transport._worker

        resource = yield _create_resource(CallerResource, worker, config)
        return RouterWebServiceRestCaller(transport, path, config, resource)


class RouterWebServiceWebhook(RouterWebService):
    """
    HTTP/POST Webhook service (part of REST-bridge).
    """
    @staticmethod
    def create(transport, path: str, config: Dict[str, Any]):
        """
        Create a new HTTP/POST Webhook service (part of REST-bridge).

        :param transport: Web transport on which to add the web service.
        :param path: HTTP path on which to add the web service.
        :param config: Web service configuration.
        :return: Web service instance.
        """
        personality = transport.worker.personality
        personality.WEB_SERVICE_CHECKERS['webhook'](personality, config)

        # the realm container on which to add the resource
        worker: Union[RouterController, ProxyController] = transport._worker

        resource = yield _create_resource(WebhookResource, worker, config)
        return RouterWebServiceWebhook(transport, path, config, resource)



##############################################################################
#
#                        Crossbar.io
#     Copyright (C) Crossbar.io Technologies GmbH. All rights reserved.
#
##############################################################################

import six
import uuid
import math
from datetime import datetime
from collections import deque

from txaio import make_logger, time_ns, perf_counter_ns

from twisted.internet.task import LoopingCall

from autobahn import util
from autobahn.wamp import message

__all__ = ('FabricRouterTrace', )

# import pyarrow as pa
# import pyarrow.parquet as pq
#
# pd.Timestamp (np.datetime64[ns])
# pa.Array.from_buffers
# pa.RecordBatch.from_arrays
# pa.Table.from_batches
# pq.ParquetFile.write_table

# a = pa.array([random.randint(0, 2**32-1) for i in range(1000)], type=pa.uint64())

# https://arrow.apache.org/docs/python/generated/pyarrow.Table.html#pyarrow.Table
# https://arrow.apache.org/docs/python/generated/pyarrow.RecordBatch.html#pyarrow.RecordBatch
# https://arrow.apache.org/docs/python/generated/pyarrow.Array.html#pyarrow.Array


class TracedMessage(object):

    __slots__ = (
        'ts',
        'pc',
        'seq',
        'realm',
        'direction',
        'session_id',
        'authid',
        'authrole',
        'msg',
    )

    def __init__(self, seq, realm, direction, session_id, authid, authrole, msg):
        self.ts = time_ns()
        self.pc = perf_counter_ns()
        self.seq = seq
        self.realm = realm
        self.direction = direction
        self.session_id = session_id
        self.authid = authid
        self.authrole = authrole
        self.msg = msg

    def marshal(self, include_message=False):
        obj = {
            u'ts': self.ts,
            u'pc': self.pc,
            u'seq': self.seq,
            u'realm': self.realm,
            u'direction': self.direction,
            u'session_id': self.session_id,
            u'authid': self.authid,
            u'authrole': self.authrole,
            u'msg_type': six.text_type(self.msg.__class__.__name__),
            u'correlation': self.msg.correlation_id,
            u'correlation_uri': self.msg.correlation_uri,
            u'correlation_is_anchor': self.msg.correlation_is_anchor,
            u'correlation_is_last': self.msg.correlation_is_last,
            u'enc_algo': self.msg.enc_algo if hasattr(self.msg, 'enc_algo') else None,
            u'enc_key': self.msg.enc_key if hasattr(self.msg, 'enc_key') else None,
            u'enc_serializer': self.msg.enc_serializer if hasattr(self.msg, 'enc_serializer') else None,
        }

        # track msg serialization sizes
        serializations = {}
        for ser, val in self.msg._serialized.items():
            serializations[ser.NAME] = len(val)
        obj[u'serializations'] = serializations

        if include_message:
            # forward raw WAMP message
            obj[u'msg'] = self.msg.marshal()

        return obj


class TracedAction(object):

    __slots__ = ('correlation_id', 'correlation_uri', 'ts', 'pc', 'seq', 'realm', 'action', 'originator', 'responders',
                 'originator_enc', 'responders_enc', 'success')

    def __init__(self, correlation_id, correlation_uri, seq, realm, action, originator, responders):
        self.correlation_id = correlation_id
        self.correlation_uri = correlation_uri
        self.ts = time_ns()
        self.pc = perf_counter_ns()
        self.seq = seq
        self.realm = realm
        self.action = action
        self.originator = originator
        self.responders = responders
        self.success = None

    def marshal(self):
        obj = {
            u'ts': self.ts,
            u'pc': self.pc,
            u'seq': self.seq,
            u'realm': self.realm,
            u'action': self.action,
            u'correlation_id': self.correlation_id,
            u'correlation_uri': self.correlation_uri,
            u'originator': self.originator,
            u'responders': self.responders,
            u'success': self.success,
        }

        return obj


class FabricRouterTrace(object):
    """

    How it works:

    A trace is always run from a router worker process. The router code calls
    into maybe_trace_rx_msg/maybe_trace_tx_msg to trace messages as they are received
    and sent to/from the router. These 2 functions check if the message is to be traced
    in the first place, and if so, create a trace record and append that to a in-memory
    list in the same (main) thread. Every 10-100ms, a looping call will trigger that
    will then take the buffered messages in the list and forward that to a background
    thread where it is written to a LMDB database file specific to this trace.
    """

    log = make_logger()

    def __init__(self,
                 session,
                 trace_id,
                 on_trace_period_finished=None,
                 trace_level=u'message',
                 trace_app_payload=False,
                 batching_period=200,
                 persist=False,
                 duration=None,
                 limit=60):
        """

        :param trace_id: The ID assigned to the trace within the router-realm.
        :type trace_id: str

        :param trace_app_payload: Flag to control tracing of the actual
            app _payload_ (args, kwargs)
        :type trace_app_payload: bool

        :param batching_period: The batching period in ms.
        :type batching_period: int

        :param persist: Flag to control trace persistence (to disk, that is LMDB).
        :type persist: bool

        :param duration: Run time of the trace in secs. If given, the trace will be automatically
            stopped after this time. Otherwise a trace needs to be stopped explicitly.
        :type duration: int

        :param limit: Limit in secs of the history kept for the trace.
        :type limit: int
        """
        self._session = session
        self._trace_id = trace_id
        self._on_trace_period_finished = on_trace_period_finished
        self._trace_level = trace_level
        self._trace_app_payload = trace_app_payload
        self._batching_period = batching_period
        self._persist = persist
        self._duration = duration
        self._limit = limit
        self._status = u'created'

        if persist:
            self._persistent_id = str(uuid.uuid4())
        else:
            self._persistent_id = None

        self._started = None
        self._ended = None

        self._seq = 0
        self._period = 0
        self._period_ts = None

        # the current batch of tracing records for trace_level==u'message'
        self._current_batch = []

        # for trace_level==u'action', map of
        # open actions: correlation => instance of TracedAction
        self._open_actions = {}

        # the accumulated data for the trace
        max_periods = int(math.ceil(float(limit) * 1000. / float(batching_period)))
        self._trace = deque(maxlen=max_periods)

        # the looping call the accumulates the current batch
        self._batch_looper = None

    def _batch_loop(self):

        period = {
            u'finished_ts': time_ns(),
            u'finished_pc': perf_counter_ns(),
            u'period': self._period,
            u'period_start': util.utcstr(self._period_ts),
        }

        current_batch = self._current_batch

        if current_batch:
            period[u'first_seq'] = current_batch[0].seq
            period[u'last_seq'] = current_batch[-1].seq
        else:
            period[u'first_seq'] = None
            period[u'last_seq'] = None

        # fire user callback with current batch
        if self._on_trace_period_finished:
            self._on_trace_period_finished(self._trace_id, period, current_batch)

        # append current batch to history
        self._trace.append((period, current_batch))

        # next period
        self._period += 1
        self._period_ts = datetime.utcnow()
        if current_batch:
            self._current_batch = []

    def start(self):
        if self._status == u'created':
            self._status = u'running'
            self._started = datetime.utcnow()
            self._period_ts = self._started
            self._batch_looper = LoopingCall(self._batch_loop)
            self._batch_looper.start(float(self._batching_period) / 1000.)
        else:
            self.log.warn('skip starting of Trace not in status "created", but "{status}"', status=self._status)

    def stop(self):
        if self._status == u'running':
            if self._batch_looper:
                if self._batch_looper.running:
                    self._batch_looper.stop()
                self._batch_looper = None
            self._status = u'stopped'
            self._ended = datetime.utcnow()
        else:
            self.log.warn('skip stopping of Trace not in status "running", but "{status}"', status=self._status)

    def marshal(self):
        if self._started:
            if self._ended:
                runtime = (self._ended - self._started).total_seconds()
            else:
                runtime = (datetime.utcnow() - self._started).total_seconds()
        else:
            runtime = None

        data = {
            u'id': self._trace_id,
            u'node_id': self._session._node_id,
            u'worker_id': self._session._worker_id,
            u'persistent_id': self._persistent_id,
            u'status': self._status,
            u'started': util.utcstr(self._started) if self._started else None,
            u'ended': util.utcstr(self._ended) if self._ended else None,
            u'runtime': runtime,
            u'options': {
                u'trace_level': self._trace_level,
                u'trace_app_payload': self._trace_app_payload,
                u'batching_period': self._batching_period,
                u'persist': self._persist,
                u'duration': self._duration,
                u'limit': self._limit
            },
            u'next_period': self._period,
            u'next_seq': self._seq,
        }
        return data

    def get_data(self, from_seq, to_seq, limit):
        # FIXME: implement selection for sequence range / limit
        res = []
        for period, batch in self._trace:
            if batch:
                if self._trace_level == u'message':
                    res.extend([trace_record.marshal(self._trace_app_payload) for trace_record in batch])
                elif self._trace_level == u'action':
                    res.extend([traced_action.marshal() for traced_action in batch])
                else:
                    raise Exception('logic error')
            if len(res) > limit:
                res = res[:limit]
                break
        return res

    def maybe_trace_rx_msg(self, session, msg):
        self._maybe_trace_msg(session, msg, u'rx')

    def maybe_trace_tx_msg(self, session, msg):
        self._maybe_trace_msg(session, msg, u'tx')

    def _maybe_trace_msg(self, session, msg, direction):
        # FIXME: implement tracing filters
        is_traced = self._status == u'running'

        if is_traced:
            self.log.debug('{direction}: {msg}', direction=direction.upper(), msg=msg)

            if self._trace_level == u'message':
                trace_record = TracedMessage(self._seq, session._realm, direction, session._session_id,
                                             session._authid, session._authrole, msg)

                self._current_batch.append(trace_record)
                self._seq += 1

            elif self._trace_level == u'action':
                if msg.correlation_is_anchor:

                    # RPC/PubSub related actions
                    if isinstance(msg, message.Call) or \
                       isinstance(msg, message.Register) or \
                       isinstance(msg, message.Unregister) or \
                       isinstance(msg, message.Publish) or \
                       isinstance(msg, message.Subscribe) or \
                       isinstance(msg, message.Unsubscribe):
                        _action = six.text_type(msg.__class__.__name__)
                    else:
                        _action = None

                    if _action:
                        traced_action = TracedAction(msg.correlation_id, msg.correlation_uri, self._seq,
                                                     session._realm, _action, session._session_id, [])

                        self._open_actions[msg.correlation_id] = traced_action
                        self._seq += 1
                        self.log.debug('New TRACE ACTION: {traced_action}', traced_action=traced_action)

                if isinstance(msg, message.Invocation) or isinstance(msg, message.Event):
                    if msg.correlation_id in self._open_actions:
                        response = {
                            'session_id': session._session_id,
                            'authid': session._authid,
                            'authrole': session._authid,
                            'enc_algo': msg.enc_algo,
                            'enc_key': msg.enc_key,
                            'enc_serializer': msg.enc_serializer,
                        }
                        self._open_actions[msg.correlation_id].responders.append(response)

                if msg.correlation_is_last:
                    if msg.correlation_id in self._open_actions:
                        traced_action = self._open_actions[msg.correlation_id]
                        traced_action.success = not isinstance(msg, message.Error)
                        del self._open_actions[msg.correlation_id]
                        self._current_batch.append(traced_action)
                        self.log.debug('TRACE ACTION finished: {traced_action}', traced_action=traced_action)

            else:
                raise Exception('internal error: invalid trace level "{}"'.format(self._trace_level))



##############################################################################
#
#                        Crossbar.io
#     Copyright (C) Crossbar.io Technologies GmbH. All rights reserved.
#
##############################################################################

import psutil

from crossbar.edge.worker.monitor._base import Monitor
from txaio import perf_counter_ns

__all__ = ('HWMonitor', )


class HWMonitor(Monitor):
    """
    Hardware monitoring. This monitor is reading hardware sensor data via psutil.
    """

    ID = u'hardware'

    def poll(self):
        """
        Measure current stats value and return new stats.
        """
        hdata = Monitor.poll(self)

        start = perf_counter_ns()

        battery = None
        bat = psutil.sensors_battery()
        if bat:
            battery = bat._asdict()

        fans = []
        for key, val in (psutil.sensors_fans() or {}).items():
            for item in val:
                item = item._asdict()
                item['device'] = key
                fans.append(item)

        temperatures = []
        for key, val in (psutil.sensors_temperatures() or {}).items():
            for item in val:
                item = item._asdict()
                item['device'] = key
                temperatures.append(item)

        hdata['battery'] = battery
        hdata['fans'] = fans
        hdata['temperatures'] = temperatures

        hdata[u'elapsed'] = perf_counter_ns() - start

        self._last_value = hdata

        return hdata



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################
"""
Tools for generating multipart requests.
"""

from uuid import uuid4

_boundary_segment = b"---------------------------"


class _Part(object):
    """
    A multipart "part".
    """
    def __init__(self, name, content, filename=None, content_type=None):

        self._content = content
        self._name = name
        self._content_type = content_type
        self._filename = filename

    def render(self):

        end_content = []
        end_content.append(b'Content-Disposition: form-data; name="')
        end_content.append(self._name)
        end_content.append(b'"')

        if self._filename:
            end_content.append(b'; filename="')
            end_content.append(self._filename)
            end_content.append(b'"')

        end_content.append(b'\r\n')

        if self._content_type:
            end_content.append(b"Content-Type: ")
            end_content.append(self._content_type)
            end_content.append(b"\r\n")

        end_content.append(b"\r\n")
        end_content.append(self._content)
        end_content.append(b"\r\n")

        return b"".join(end_content)


class Multipart(object):
    """
    A multipart request.
    """
    def __init__(self):
        self._parts = []
        self._boundary = uuid4().hex.encode('ascii')

    def add_part(self, name, content, filename=None, content_type=None):

        self._parts.append(_Part(name, content, content_type=content_type, filename=filename))

    def render(self):
        """
        Render this Multipart into content and the requisite headers.
        """
        boundary = _boundary_segment + self._boundary

        end_content = []
        end_content.append(b"--" + boundary + b"\r\n")
        end_content.append((b"--" + boundary + b"\r\n").join([part.render() for part in self._parts]))
        end_content.append(b"--" + boundary)
        end_content.append(b"--\r\n")
        content = b"".join(end_content)

        headers = {
            b"content-type": [b"multipart/form-data; boundary=" + boundary],
            b"content-length": [str(len(content)).encode('ascii')]
        }

        return (content, headers)



##############################################################################
#
#                        Crossbar.io
#     Copyright (C) Crossbar.io Technologies GmbH. All rights reserved.
#
##############################################################################

from twisted.internet.defer import succeed
from autobahn.util import utcnow

from crossbar.edge.worker.monitor._base import Monitor

__all__ = ('MemoryMonitor', )


class MemoryMonitor(Monitor):
    """
    RAM monitoring.
    """

    ID = u'memory'

    def __init__(self, config=None):
        Monitor.__init__(self, config)

    def poll(self):
        """
        Measure current stats value and return new stats.
        """
        Monitor.poll(self)

        # create new, empty event
        #
        current = {
            # the UTC timestamp when measurement was taken
            u'timestamp': utcnow(),

            # the effective last period in secods
            u'last_period': self._last_period,
        }

        # FIXME: add ratio of ram usage

        with open("/proc/meminfo") as f:
            res = f.read()
            new = res.split()
            new_clean = [x.replace(":", "") for x in new if x != 'kB']
            for i in range(0, len(new_clean), 2):
                k = u'{}'.format(new_clean[i])
                current[k] = int(new_clean[i + 1])

        self._last_value = current

        return succeed(self._last_value)
